##
# File:    FormatUtils.py
# Date:    10-April-2010
#
# Updates:
# 23-April-2010  jdw Add rcsb internal cif to pdbx
# 18-April-2013  jdw update to latest RcsbDbUtility() class -
# 29-May  -2013  jdw Add pdb2pdbxDepositOp()  and pdbx2pdbxDepositOp() methods
# 15-Aug  -2013  jdw Add mtz2pdbxOp()
# 31-Dec  -2013  jdw add timeout for mtz2pdbxOp()
# 16-Jan  -2014  jdw repoint dst3 in mtz2pdbxOp() to logfile
# 26-Sep  -2014  jdw add pdbx2nmrstarAnnotOp() from annotation package
##
"""
Module of format translation utility operations supporting the call protocol of the ProcessRunner() class.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import os
import sys
import traceback

from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase
from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility
from wwpdb.utils.wf.plugins.SFConvert import SFConvert


class FormatUtils(UtilsBase):

    """Utility class to perform file format conversions.

    Current supported operations include:
    - annot-pdbx2pdb (cif2pdb)
    - annot-pdb2pdbx (pdb2cif)
    - annot-rcsb2pdbx (cif2pdbx)
    - annot-cif2cif-dep   (pdbx2pdbx)
    - annot-pdb2cif-dep   (pdb2cif)

    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=False, log=sys.stderr):
        super(FormatUtils, self).__init__(verbose, log)
        self.__cleanUp = True
        """Flag to remove any temporary directories created by this class.
        """
        #

    def pdbx2pdbOp(self, **kwArgs):
        """Performs PDBx(cif) to PDB format conversion operation."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            pdbPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.op("annot-cif2pdb")
            dp.exp(pdbPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+FormatUtils.pdbx2pdbOp() - PDB  file path: %s\n" % pdbPath)
                self._lfh.write("+FormatUtils.pdbx2pdbOp() - PDBx file path: %s\n" % pdbxPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def pdb2pdbxOp(self, **kwArgs):
        """Perform PDB to PDBx(cif) format conversion operation."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbPath = inpObjD["src"].getFilePathReference()
            pdbxPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbPath)
            dp.op("annot-pdb2cif")
            dp.exp(pdbxPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+FormatUtils.pdb2pdbxOp() - PDB  file path: %s\n" % pdbPath)
                self._lfh.write("+FormatUtils.pdb2pdbxOp() - PDBx file path: %s\n" % pdbxPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def pdb2pdbxDepositOp(self, **kwArgs):
        """Perform PDB to PDBx(cif) format conversion operation (special processing for deposition sessions)"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbPath = inpObjD["src"].getFilePathReference()
            pdbxPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbPath)
            dp.op("annot-pdb2cif-dep")
            dp.exp(pdbxPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+FormatUtils.pdb2pdbxDepositOp() - PDB  file path: %s\n" % pdbPath)
                self._lfh.write("+FormatUtils.pdb2pdbxDepositOp() - PDBx file path: %s\n" % pdbxPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def rcsb2pdbxOp(self, **kwArgs):
        """Perform RCSB(cif) to PDBx(cif) format conversion operation."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            rcsbPath = inpObjD["src"].getFilePathReference()
            pdbxPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(rcsbPath)
            dp.op("annot-cif2cif")
            dp.exp(pdbxPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+FormatUtils.cif2cifOp() - RCSB cif file path: %s\n" % rcsbPath)
                self._lfh.write("+FormatUtils.cif2cifOp() - PDBx file path: %s\n" % pdbxPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def pdbx2pdbxDepositOp(self, **kwArgs):
        """Perform  PDBx(cif) to PDBx(cif) format conversion operation (special processing deposition sessions)"""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            rcsbPath = inpObjD["src"].getFilePathReference()
            pdbxPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(rcsbPath)
            dp.op("annot-cif2cif-dep")
            dp.exp(pdbxPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+FormatUtils.pdbx2pdbxDepositOp() - PDBx input  file path: %s\n" % rcsbPath)
                self._lfh.write("+FormatUtils.pdbx2pdbxDepositOp() - PDBx output file path: %s\n" % pdbxPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def mtz2pdbxOp(self, **kwArgs):
        """Perform  MTZ or other SF file  to  PDBx(cif) format conversion operation."""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            sfPath = inpObjD["src1"].getFilePathReference()
            sfFmt = inpObjD["src1"].getFileFormat()
            xyzPath = inpObjD["src2"].getFilePathReference()
            #
            #
            sfPdbxFilePath = outObjD["dst1"].getFilePathReference()
            sfDiagFilePath = outObjD["dst2"].getFilePathReference()
            logFilePath = outObjD["dst3"].getFilePathReference()
            #
            dirPath = outObjD["dst1"].getDirPathReference()
            #
            # logPath=os.path.join(dirPath,"mtz2pdbx.log")
            dmpPath = os.path.join(dirPath, "mtzdmp.log")
            #
            timeout = int(uD["timeout"])
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")

            if sfFmt == "pdbx":
                # See if we can handle specially
                sfc = SFConvert()
                if sfc.isSpecialSF(sfPath):
                    ok = sfc.handleSpecialSF(sfPath, sfPdbxFilePath, xyzPath, sfDiagFilePath, logFilePath)
                    self._lfh.write("Special handling of SF file returns %s\n" % ok)
                    if ok is True:
                        return True
                    # Else fall through

            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(sfPath)
            if (xyzPath is not None) and os.path.exists(xyzPath):
                dp.addInput(name="xyz_file_path", value=xyzPath, type="file")
            if timeout > 0:
                dp.setTimeout(timeout)
            dp.op("annot-sf-convert")
            dp.expLog(logFilePath)
            dp.expList(dstPathList=[sfPdbxFilePath, sfDiagFilePath, dmpPath])

            #
            myStatus = self.__checkMergeStatus(logFilePath)
            outObjD["dst4"].setValue(myStatus)

            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+FormatUtils.mtz2pdbxOp() - timeout :      %d\n" % timeout)
                self._lfh.write("+FormatUtils.mtz2pdbxOp() - SF  input  file path:      %s\n" % sfPath)
                self._lfh.write("+FormatUtils.mtz2pdbxOp() - XYZ input  file path:      %s\n" % xyzPath)
                self._lfh.write("+FormatUtils.mtz2pdbxOp() - PDBx SF output file path:  %s\n" % sfPdbxFilePath)
                self._lfh.write("+FormatUtils.mtz2pdbxOp() - Log file path:             %s\n" % logFilePath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def nmrstar2pdbxOp(self, **kwArgs):
        """Perform NMRSTAR to PDBx(cif) format conversion operation."""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            strPath = inpObjD["src"].getFilePathReference()
            pdbxPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            dId = uD["data_set_id"]
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(strPath)
            dp.addInput(name="data_set_id", value=dId)
            dp.op("annot-nmrstar2pdbx")
            dp.exp(pdbxPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+FormatUtils.nmrstar2pdbxOp() - NMRSTAR  file path: %s\n" % strPath)
                self._lfh.write("+FormatUtils.nmrstar2pdbxOp() - PDBx file path:     %s\n" % pdbxPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def pdbx2nmrstarAnnotOp(self, **kwArgs):
        """Perform PDBx(cif) to NMRSTAR format conversion operation.  (using annotation package stack)"""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = outObjD["src"].getFilePathReference()
            strPath = inpObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            dId = uD["data_set_id"]
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(pdbxPath)
            dp.addInput(name="pdb_id", value=dId)
            dp.op("annot-pdbx2nmrstar")
            dp.exp(strPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+FormatUtils.pdbx2nmrstarAnnotOp() - NMRSTAR  file path: %s\n" % strPath)
                self._lfh.write("+FormatUtils.pdbx2nmrstarAnnotOp() - PDBx file path:     %s\n" % pdbxPath)
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def __checkMergeStatus(self, logFilePath):
        status = "ok"
        if os.access(logFilePath, os.R_OK):
            ifh = open(logFilePath, "r")
            for line in ifh:
                if str(line).upper().startswith("++ERROR") or str(line).upper().startswith("ERROR:"):
                    return "error"
                if str(line).upper().startswith("++WARN") or str(line).upper().startswith("WARN:"):
                    return "warn"
            ifh.close()
        else:
            return "error"
        return status



##
# File:    PrdSearchUtils.py
# Date:    10-Sept-2010
#
# Updates:
# 2013-03-15    ZF   Explicitly setting name of working directory used during calls for "prd-search" processing.
##
"""
Module of entity transformer utility operations supporting the call protocol of the ProcessRunner() class.

"""
__docformat__ = "restructuredtext en"
__author__ = "Zukang Feng"
__email__ = "zfeng@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import os
import sys
import traceback

from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase
from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility


class PrdSearchUtils(UtilsBase):
    """Utility class to perform PRD search operation.

    The method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    The method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=False, log=sys.stderr):
        super(PrdSearchUtils, self).__init__(verbose, log)
        self.__cleanUp = False
        """Flag to remove any temporary directories created by this class.
        """
        #

    def PrdSearchOp(self, **kwArgs):
        """Performs PRD search on PDBx format files."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            resultFilePath = outObjD["dst"].getFilePathReference()
            #
            WorkingDirPath = os.path.join(dirPath, "search")
            firstModelPath = os.path.join(WorkingDirPath, "firstmodel.cif")
            logFilePath = os.path.join(WorkingDirPath, "search-prd.log")
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.setWorkingDir(WorkingDirPath)
            dp.addInput(name="firstmodel", value=firstModelPath)
            dp.addInput(name="logfile", value=logFilePath)
            dp.imp(pdbxPath)
            dp.op("prd-search")
            dp.exp(resultFilePath)
            if self.__cleanUp:
                dp.cleanup()
            if not os.access(resultFilePath, os.R_OK):
                return False
            #
            logFilePath = os.path.join(WorkingDirPath, "process-prd.log")
            dp1 = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp1.addInput(name="resultFile", value=resultFilePath)
            dp1.addInput(name="logfile", value=logFilePath)
            dp1.op("prd-process-summary")
            if self.__cleanUp:
                dp1.cleanup()
            # """
            # myReqObj = InputRequest({}, verbose=self._verbose, log=self._lfh)
            # myReqObj.setValue("TopSessionPath", cI.get('SITE_WEB_APPS_TOP_SESSIONS_PATH'))
            # myReqObj.setValue("TopPath", cI.get('SITE_WEB_APPS_TOP_PATH'))
            # myReqObj.setValue("WWPDB_SITE_ID",  siteId)
            # prdUtil = ProcessPrdSummary(reqObj=myReqObj, verbose=self._verbose, log=self._lfh)
            # prdUtil.setTopDirPath(dirPath)
            # prdUtil.setPrdSummaryFile(resultFilePath)
            # prdUtil.run()
            # #
            # """
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False



##
# File:    ActionRegistry.py
# Date:    1-April-2010
#
# Updates:
#
# 21-Apr-2010 jdw Added method getActions()
#  7-Sep-2015 jdw Return names in parse order -
##
"""
Repository for process action definitions.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import traceback
from wwpdb.utils.wf.process.ActionRegistryIo import ActionRegistryIo
from wwpdb.utils.config.ConfigInfo import ConfigInfo


class ActionRegistry(object):

    """Container and manager class for the registry of supported process actions.

    The action registry has the following internal data orgaization:

    - A dictionary with unique action identifier referencing a dictionary with the following content:

      + INPUT_INFO_LIST & OUTPUT_INFO_LIST,  list of required features of the input and out
        data objects stored as tuples of (reference_type, (data_type,container_type), selector_type)
      + USER_PARAMETER_DICTIONARY,  container for user settable parameters passed to
        the action method.
      + INTERNAL_PARAMETER_DICTIONARY,  container for internal parameters passed to
        the action method.
      + MODULE_NAME,   Python module(class) name containing the target method
      + METHOD_NAME,   Python method name

    Method prototype:

    method(inputObjList=[],outputObjList=[],parameterDictionary={})

    """

    def __init__(self):
        self.__cI = ConfigInfo()
        regPath = self.__cI.get("SITE_REGISTRY_FILE_PATH")
        aR = ActionRegistryIo(filePath=regPath)
        self.__D = aR.getRegistry()

    def getActions(self):
        """Returns:

        List of action identifiers defined in the action registry. An empty list is
        returned if the an exception is encountered.

        """
        try:
            return self.__D.keys()
        except Exception as _e:  # noqa: F841
            return []

    def isDefinedAction(self, actionId):
        """Returns:

        True if the input action identifier is defined in the registry or False otherwise.
        """
        if actionId in self.__D:
            return True
        else:
            return False

    def getMethodName(self, actionId):
        """Returns:

        The Python method name associated with the input action identifier
        or None if the action is not defined.
        """
        try:
            return self.__D[actionId]["METHOD_NAME"]
        except Exception as _e:  # noqa: F841
            return None

    def getModuleName(self, actionId):
        """Returns:

        The Python module name containing the method associated with the input action identifier
        or None if the action is not defined.
        """

        try:
            return self.__D[actionId]["MODULE_NAME"]
        except Exception as _e:  # noqa: F841
            return None

    def getUserParameterDict(self, actionId):
        """Returns:

        The user parameter dictionary for the input action identifier
        or {} if the action is not defined.
        """

        try:
            return self.__D[actionId]["USER_PARAMETER_DICT"]
        except Exception as _e:  # noqa: F841
            return {}

    def getInternalParameterDict(self, actionId):
        """Returns:

        The internal parameter dictionary for the input action identifier
        or {} if the action is not defined.
        """

        try:
            return self.__D[actionId]["INTERNAL_PARAMETER_DICT"]
        except Exception as _e:  # noqa: F841
            return {}

    def getInputObjectCount(self, actionId):
        """Returns:

        The number of input data objects required for input action identifier
        or 0 if the action is not defined.
        """

        try:
            return len(self.__D[actionId]["INPUT_INFO_LIST"].keys())
        except Exception as _e:  # noqa: F841:
            traceback.print_exc(file=sys.stderr)
            return 0

    def getInputObjectNames(self, actionId):
        """Returns:

        The list of names of input data objects required for input action identifier
        or [] if the action is not defined.
        """

        try:
            return self.__D[actionId]["INPUT_NAME_LIST"]
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=sys.stderr)
            return []

    def getOutputObjectCount(self, actionId):
        """Returns:

        The number of output data objects required for the input action identifier
        or 0 if the action is not defined.
        """

        try:
            return len(self.__D[actionId]["OUTPUT_INFO_LIST"].keys())
        except Exception as _e:  # noqa: F841
            return 0

    def getOutputObjectNames(self, actionId):
        """Returns:

        The list of names of output data objects required for the input action identifier
        or [] if the action is not defined.
        """

        try:
            return self.__D[actionId]["OUTPUT_NAME_LIST"]
        except Exception as _e:  # noqa: F841
            return []

    def setUserParameter(self, actionId, paramKey, paramValue):
        """Set a user adjustable parameter for the input action.  The parameter is
        identifiers by *paramKey* and the value is specified as *paramValue*.

        Returns:

        True is the parameter setting was succesful or False otherwise.

        """
        try:
            self.__D[actionId]["USER_PARAMETER_DICT"][paramKey] = paramValue
            return True
        except Exception as _e:  # noqa: F841
            return False

    def getUserParameter(self, actionId, paramKey):
        """Returns:

        The value of a user adjustable parameter for the input action.  The
        user parameteris identified by the *paramKey*.

        """
        try:
            return self.__D[actionId]["USER_PARAMETER_DICT"][paramKey]
        except Exception as _e:  # noqa: F841
            return None

    def isSetInputReferenceType(self, actionId, inpName="1"):
        """Is the reference type for the specified action and input object assigned?

        Returns:

        True if the reference type is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["dataReferenceType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getInputReferenceType(self, actionId, inpName="1"):
        """Get the reference type for the specified action and input object.

        Returns:

        The reference type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["dataReferenceType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetInputContainerType(self, actionId, inpName="1"):
        """Is the container type for the specified action and input object assigned?

        Returns:

        True if the container type is assigned or False otherwise.

        """

        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["containerType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getInputContainerType(self, actionId, inpName="1"):
        """Get the container type for the specified action and input object.

        Returns:

        The container type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["containerType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetInputValueType(self, actionId, inpName="1"):
        """Is the value type for the specified action and input object assigned?

        Returns:

        True if the value type is assigned or False otherwise.

        """

        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["containerType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getInputValueType(self, actionId, inpName="1"):
        """Get the value type for the specified action and input object.

        Returns:

        The value type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["valueType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetInputSelectorType(self, actionId, inpName="1"):
        """Is the selector type for the specified action and input object assigned?

        Returns:

        True if the selector type is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["selectorType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getInputSelectorType(self, actionId, inpName="1"):
        """Get the selector type for the specified action and input object.

        Returns:

        The selector type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["selectorType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetInputContentType(self, actionId, inpName="1"):
        """Is the content type for the specified action and input object assigned?

        Returns:

        True if the content type is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["contentType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getInputContentType(self, actionId, inpName="1"):
        """Get the content type for the specified action and input object.

        Returns:

        The content type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["contentType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetInputFileFormat(self, actionId, inpName="1"):
        """Is the file format for the specified action and input object assigned?

        Returns:

        True if the file format is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["fileFormat"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getInputFileFormat(self, actionId, inpName="1"):
        """Get the file format for the specified action and input object.

        Returns:

        The file format if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["INPUT_INFO_LIST"][inpName]["fileFormat"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetOutputReferenceType(self, actionId, inpName="1"):
        """Is the reference type for the specified action and output object assigned?

        Returns:

        True if the reference type is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["dataReferenceType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getOutputReferenceType(self, actionId, inpName="1"):
        """Get the reference type for the specified action and output object.

        Returns:

        The reference type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["dataReferenceType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetOutputContainerType(self, actionId, inpName="1"):
        """Is the container type for the specified action and output object assigned?

        Returns:

        True if the container type is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["containerType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getOutputContainerType(self, actionId, inpName="1"):
        """Get the container type for the specified action and output object.

        Returns:

        The container type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["containerType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetOutputValueType(self, actionId, inpName="1"):
        """Is the value type for the specified action and output object assigned?

        Returns:

        True if the value type is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["valueType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getOutputValueType(self, actionId, inpName="1"):
        """Get the value type for the specified action and output object.

        Returns:

        The value type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["valueType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetOutputSelectorType(self, actionId, inpName="1"):
        """Is the selector type for the specified action and output object assigned?

        Returns:

        True if the selector type is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["selectorType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getOutputSelectorType(self, actionId, inpName="1"):
        """Get the selector type for the specified action and output object.

        Returns:

        The selector type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["selectorType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetOutputContentType(self, actionId, inpName="1"):
        """Is the content type for the specified action and output object assigned?

        Returns:

        True if the content type is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["contentType"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getOutputContentType(self, actionId, inpName="1"):
        """Get the content type for the specified action and output object.

        Returns:

        The content type if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["contentType"]
        except Exception as _e:  # noqa: F841
            return None

    def isSetOutputFileFormat(self, actionId, inpName="1"):
        """Is the file format for the specified action and output object assigned?

        Returns:

        True if the file format is assigned or False otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["fileFormat"] is not None
        except Exception as _e:  # noqa: F841
            return False

    def getOutputFileFormat(self, actionId, inpName="1"):
        """Get the file format for the specified action and output object.

        Returns:

        The file format if this is defined or None otherwise.

        """
        try:
            return self.__D[actionId]["OUTPUT_INFO_LIST"][inpName]["fileFormat"]
        except Exception as _e:  # noqa: F841
            return None



##
# File:    DataValueContainer.py
# Date:    28-Mar-2010
#
# Updates:
#
##
"""
Container for data values.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

from datetime import datetime, date

# For python 2/3 compatible comparsigon with isinstace
from builtins import str


class DataValueContainer(object):

    """Container for data values.

    Supported container types include:
    - individual bool, int, float, string, date, or datetime values
    - lists of bool, int, float, string,  date, datetime values

    """

    def __init__(self):
        super(DataValueContainer, self).__init__()
        # sys.stderr.write("DataValueContainer.__init_()\n")
        #
        self.__valueTypeName = None
        """ A supported data type name such:
            - boolean
            - int or integer
            - float or double
            - string
            - date
            - datetime
            """
        #
        self.__valueType = type(None)
        """ Intrinsic Python type corresponding to the valueTypeName.

            Types are defined in the Python class `types`.
        """
        #
        self.__containerTypeName = None
        """ Container type setting:
            - value, a single value
            - list,  a list/vector of values
            - dict,  a dictionary of key/values pairs

        """
        #
        self.__value = None

    def isValueValid(self):
        """Performs a sanity type check on the current value and container types.

        Returns:

        True if value and container types correspond to the current type settings or False otherwise.
        """
        if self.__containerTypeName == "list":
            if isinstance(self.__value, list):
                for v in self.__value:
                    if not isinstance(v, self.__valueType):
                        return False
                return True
            else:
                return False
        elif self.__containerTypeName == "dict":
            if isinstance(self.__value, dict):
                return True
            else:
                return False

        else:
            if isinstance(self.__value, self.__valueType):
                return True
            else:
                return False

    def isValueSet(self):
        """Performs a check if the current data value has been set.

        Returns:

        True if the value has been set or False otherwise.

        """
        return self.__value is not None

    def setValue(self, value):
        self.__value = value

    def getValue(self):
        return self.__value

    def setValueTypeName(self, typeName):
        """Set the data type name for the container.

        Supported types include:
        - boolean
        - int or integer
        - float or double
        - string
        - date
        - datetime

        Returns:

        True if the input typeName is a supported type or False otherwise.
        """
        if str(typeName) in ["boolean", "int", "integer", "float", "double", "string", "date", "datetime"]:
            self.__valueTypeName = str(typeName)
            if typeName == "bool" or typeName == "boolean":
                self.__valueType = bool
            elif typeName == "integer" or typeName == "int":
                self.__valueType = int
            elif typeName == "float" or typeName == "double":
                self.__valueType = int
            elif typeName == "string":
                self.__valueType = str
            elif typeName == "date":
                tt = date(2010, 1, 1)
                self.__valueType = tt.__class__
            elif typeName == "datetime":
                tt = datetime(2010, 1, 1)
                self.__valueType = tt.__class__
            else:
                return False
            return True
        else:
            return False

    def setContainerTypeName(self, containerName):
        """Set the container type name.

        Supported container types include:
        - value,  a individual value
        - list,   list or vector of values
        - dict,   dictionary of key,value pairs

        Returns:

        True for supported container types or False otherwise.

        """
        if containerName in ["value", "list", "dict"]:
            self.__containerTypeName = containerName
            return True
        else:
            return False

    def getContainerTypeName(self):
        return self.__containerTypeName

    def getValueTypeName(self):
        return self.__valueTypeName



##
# File:    SeqdbUtils.py
# Date:    25-April-2010
#
# Updates:
#    22-Mar-2013 jdw overhaul for version 2 sequence module
#    22-Mar-2013 jdw provide matchAppOp() to search all relevant entities in a single method.
#    25-Mar-2013 jdw respect the destination setting when returning files in match-all
#    20-Apr-2014 jdw add additional export of polymer linkage report
#    22-May-2014 jdw add perform and store alignments with default selections
##
"""
Module of sequence database search utilities.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import os
import sys
import traceback
from wwpdb.utils.config.ConfigInfo import getSiteId
from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase

try:
    # We will have present on annotation system - but allow testing
    from wwpdb.apps.seqmodule.webapp.SeqModWebRequest import SeqModInputRequest
except ImportError:
    pass

try:
    # We will have present on annotation system - but allow testing
    from wwpdb.apps.seqmodule.control.DataImporter import DataImporter
except ImportError:
    pass

from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility


class SeqdbUtils(UtilsBase):
    """Utility class to perform sequence database searches.

    Current supported operations include:
    - find matching sequences for all polymer entities in a model file
    - find matching sequences of a specific polymer identified by entity_id

    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=True, log=sys.stderr):
        super(SeqdbUtils, self).__init__(verbose, log)
        self.__cleanUp = False
        """Flag to remove any temporary directories created by this class.
        """
        # Init here to keep pylint happy
        self.__maxRefAlign = 0
        self.__siteId = None
        self.__reqObj = None
        # self.__sessionId = None
        # self.__sessionObj = None
        # self.__sessionPath = None
        self.__doAutoProcessFlag = False
        self.__includeSeqAssignFileFlag = False
        #

    def mySetup(self, topSessionPath="."):
        """Setup application environment for managing session storage of
        temporaty data files.
        """
        self.__maxRefAlign = 100
        self.__siteId = getSiteId(defaultSiteId="WWPDB_DEPOLY_TEST")
        #
        self.__reqObj = SeqModInputRequest({}, verbose=self._verbose, log=self._lfh)
        self.__reqObj.setValue("TopSessionPath", topSessionPath)
        self.__reqObj.setValue("WWPDB_SITE_ID", self.__siteId)
        #
        #  self.__sessionId = self.__reqObj.getSessionId()
        # The following has side effects
        _sessionObj = self.__reqObj.newSessionObj()  # noqa: F841
        #  self.__sessionPath = self.__sessionObj.getPath()

    def matchAllOp(self, **kwArgs):
        """Find matching sequences for all entities."""
        try:
            self.__doAutoProcessFlag = False
            self.__includeSeqAssignFileFlag = False
            self.__runMatchAllOp(kwArgs, "matchAllOp")
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False
        #

    def matchAllAutoOp(self, **kwArgs):
        """Find matching sequences for all entities."""
        try:
            self.__doAutoProcessFlag = True
            self.__includeSeqAssignFileFlag = True
            self.__runMatchAllOp(kwArgs, "matchAllAutoOp")
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False
        #

    def matchEntityOp(self, **kwArgs):
        """Find matching sequences for all entities."""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            dstPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            entityId = str(uD["entity_id"])
            #
            # details of the input file -
            #
            depDataSetId = inpObjD["src"].getDepositionDataSetId()
            instanceId = inpObjD["src"].getWorkflowInstanceId()
            fileSource = inpObjD["src"].getStorageType()
            #
            self.mySetup(topSessionPath=dirPath)
            self.__reqObj.setValue("identifier", depDataSetId)
            self.__reqObj.setValue("instance", instanceId)
            #
            dI = DataImporter(reqObj=self.__reqObj, fileSource=fileSource, maxRefAlign=self.__maxRefAlign, verbose=self._verbose, log=self._lfh)
            dI.copyModelFile(inputFileSource=fileSource, inputWfInstanceId=instanceId)
            dI.copyFiles(messageHead="SeqdbUtils.matchAllOp(OnStart)")
            entityIdList, ok = dI.loadSeqDataAssemble(selectedEntityIdList=[entityId], doRefSearch=True)
            #
            # Return the files according to the destination setting --
            #
            instanceId = outObjD["dst"].getWorkflowInstanceId()
            fileSource = outObjD["dst"].getStorageType()
            dI.copyFiles(
                inputFileSource="session",
                outputFileSource=fileSource,
                outputWfInstanceId=instanceId,
                versionIndex=4,
                includePolyLinkFile=True,
                entityIdList=entityIdList,
                messageHead="SeqdbUtils.matchEntityOp(OnFinish)",
            )
            #
            if self._verbose:
                self._lfh.write("+SeqdbUtils.matchEntityOp() - Input model PDBx file path: %s\n" % pdbxPath)
                self._lfh.write("+SeqdbUtils.matchEntityOp() - Entity id : %s\n" % entityId)
                self._lfh.write("+SeqdbUtils.matchEntityOp() - Output result path: %s\n" % dstPath)
            #
            if self.__cleanUp:
                pass
            #
            return ok
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def updateModelWithSeqAssignmentOp(self, **kwArgs):
        """Performs sequence assignment update operation on PDBx format model file."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src1"].getFilePathReference()
            depDataSetId = inpObjD["src1"].getDepositionDataSetId()
            instanceId = inpObjD["src1"].getWorkflowInstanceId()
            fileSource = inpObjD["src1"].getStorageType()
            #
            seqAssignFilePath = inpObjD["src2"].getFilePathReference()
            #
            outputModelPdbxPath = outObjD["dst"].getFilePathReference()
            dirPath = outObjD["dst"].getDirPathReference()
            #
            self.mySetup(topSessionPath=dirPath)
            self.__reqObj.setValue("identifier", depDataSetId)
            self.__reqObj.setValue("instance", instanceId)
            #
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=self.__siteId, verbose=self._verbose, log=self._lfh)
            #
            if (seqAssignFilePath is not None) and os.path.exists(seqAssignFilePath):
                dp.addInput(name="seqmod_assign_file_path", value=seqAssignFilePath, type="file")
            #
            dp.imp(pdbxPath)
            dp.op("annot-merge-sequence-data")
            dp.exp(outputModelPdbxPath)
            if self.__cleanUp:
                dp.cleanup()
            if self._verbose:
                self._lfh.write("+SeqdbUtils.updateModelWithSeqAssignmentOp() - PDBx file path:        %s\n" % pdbxPath)
                self._lfh.write("+SeqdbUtils.updateModelWithSeqAssignmentOp() - Seq assign file path:  %s\n" % seqAssignFilePath)
                self._lfh.write("+SeqdbUtils.updateModelWithSeqAssignmentOp() - PDBx output file path: %s\n" % outputModelPdbxPath)
            #
            # Copy files back to archive
            #
            dI = DataImporter(reqObj=self.__reqObj, fileSource=fileSource, maxRefAlign=self.__maxRefAlign, verbose=self._verbose, log=self._lfh)
            dI.copyModelFile(inputFileSource=fileSource, inputWfInstanceId=instanceId, outputFileSource="archive", versionIndex=4)
            dI.copyFiles(
                inputFileSource=fileSource,
                inputWfInstanceId=instanceId,
                outputFileSource="archive",
                versionIndex=4,
                includePolyLinkFile=True,
                includeSeqAssignFile=True,
                messageHead="SeqdbUtils.updateModelWithSeqAssignmentOp(OnFinish)",
            )
            #
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False
        #

    def __runMatchAllOp(self, kwArgs, functionName):
        """Find matching sequences for all entities."""
        (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
        pdbxPath = inpObjD["src"].getFilePathReference()
        dirPath = outObjD["dst"].getDirPathReference()
        #
        # details of the input file -
        #
        depDataSetId = inpObjD["src"].getDepositionDataSetId()
        instanceId = inpObjD["src"].getWorkflowInstanceId()
        fileSource = inpObjD["src"].getStorageType()
        #
        self.mySetup(topSessionPath=dirPath)
        self.__reqObj.setValue("identifier", depDataSetId)
        self.__reqObj.setValue("instance", instanceId)
        #
        # Do all the work here
        #
        dI = DataImporter(reqObj=self.__reqObj, fileSource=fileSource, maxRefAlign=self.__maxRefAlign, verbose=self._verbose, log=self._lfh)
        dI.copyModelFile(inputFileSource=fileSource, inputWfInstanceId=instanceId)
        dI.copyFiles(messageHead="SeqdbUtils." + functionName + "(OnStart)")
        entityIdList, _ok = dI.loadSeqDataAssemble(doAutoProcess=self.__doAutoProcessFlag)
        #
        # Return the files according to the destination setting --
        #
        instanceId = outObjD["dst"].getWorkflowInstanceId()
        fileSource = outObjD["dst"].getStorageType()
        dI.copyFiles(
            inputFileSource="session",
            outputFileSource=fileSource,
            outputWfInstanceId=instanceId,
            versionIndex=4,
            includePolyLinkFile=True,
            includeSeqAssignFile=self.__includeSeqAssignFileFlag,
            entityIdList=entityIdList,
            messageHead="SeqdbUtils." + functionName + "(OnFinish)",
        )
        #
        if self._verbose:
            self._lfh.write("+SeqdbUtils.matchAllOp() - Input model PDBx file path: %s\n" % pdbxPath)
            self._lfh.write("+SeqdbUtils.matchAllOp() - Output dir path: %s\n" % dirPath)
        #



"""

    File:    WfDbApi.py

    Providing APIs for workflow engine and workflow manager in D&A tool

   __author__    = "Li Chen"
   __email__     = "lchen@rcsb.rutgers.edu"
   __version__   = "V0.01"
   __Date__      = "April 21, 2010"

 Updates:
  07-Feb-2014  to  add processStatus()
  07-Feb-2014  jdw add socket support
   4-Aug-2014  jdw add debug flag to minimize connection tracking logging
  26-sep-2017  ep  runInsertSQL/runUpdateSQL allow for parameterized arguments
                   to pass to db execute to handle quoting
  15-Jul-2019  ep  add siteId as optional argument to __init__

"""

import os
import sys
import time
import datetime
import MySQLdb

#
from wwpdb.utils.wf.dbapi.DbConnection import DbConnection
from wwpdb.utils.wf.dbapi.DbCommand import DbCommand
from wwpdb.utils.wf.schema.WfSchemaMap import WfSchemaMap
from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.wf.dbapi.WFEtime import getTimeNow


class WfDbApi:
    """
    Sample of WfSchemaMap class:

     _schemaMap = {
       "DEPOSITION" : {
                 "ATTRIBUTES"{dictionary of columns}
                 "TABLE_NAME"  : "deposition"
                      }
       "WF_TASK"    : {
                 "ATTRIBUTES"{dictionary of columns}
                 "TABLE_NAME"  : "wf_task"
                      }
                  }

    """

    # This could be rewritten wih inheritance - later
    # pylint is picking up on shared data among instances and thinks unused
    __schemaWf = WfSchemaMap._schemaMap  # pylint: disable=protected-access,unused-private-member
    __selectList = WfSchemaMap._selectColumns  # pylint: disable=protected-access,unused-private-member
    __constraintList = WfSchemaMap._constraintList  # pylint: disable=protected-access,unused-private-member
    # __statusList = WfSchemaMap._columnForStatus
    __statusList = WfSchemaMap._usefulItems[0:3]  # pylint: disable=protected-access,unused-private-member
    __columnList = WfSchemaMap._usefulItems  # pylint: disable=protected-access,unused-private-member
    __tableList = WfSchemaMap._tables  # pylint: disable=protected-access,unused-private-member
    __idList = WfSchemaMap._objIds  # pylint: disable=protected-access,unused-private-member
    __refList = WfSchemaMap._referencePairs  # pylint: disable=protected-access,unused-private-member
    __sqlJoinStr = WfSchemaMap._tableJoinSyntext  # pylint: disable=protected-access,unused-private-member
    __orderBy = WfSchemaMap._orderBy  # pylint: disable=protected-access,unused-private-member
    __userInfo = WfSchemaMap._userInfo  # pylint: disable=protected-access,unused-private-member

    def __init__(self, log=sys.stderr, verbose=False, siteId=None):
        """
        Either siteId needs to be specified or Environmental variable WWPDB_SITE_ID needs to be set
        for ConfigInfo() to obtain the correct details -

        """

        self.__Nretry = 5
        self.__lfh = log
        self.__verbose = verbose
        self.__debug = False
        cI = ConfigInfo(siteId=siteId)
        self.__dbServer = cI.get("SITE_DB_SERVER")
        self.__dbHost = cI.get("SITE_DB_HOST_NAME")
        self.__dbName = cI.get("SITE_DB_DATABASE_NAME")
        self.__dbUser = cI.get("SITE_DB_USER_NAME")
        self.__dbPw = cI.get("SITE_DB_PASSWORD")
        self.__dbSocket = cI.get("SITE_DB_SOCKET")
        self.__dbPort = int("%s" % cI.get("SITE_DB_PORT_NUMBER"))

        if self.__debug:
            self.__lfh.write("\n+WfDbApi.__init__() using socket %r\n" % self.__dbSocket)
            self.__lfh.write("+WfDbApi.__init__() using socket environment reference %r\n" % os.getenv("SITE_DB_SOCKET", None))

        self.__myDb = DbConnection(
            dbServer=self.__dbServer, dbHost=self.__dbHost, dbName=self.__dbName, dbUser=self.__dbUser, dbPw=self.__dbPw, dbPort=self.__dbPort, dbSocket=self.__dbSocket
        )

        self.__dbcon = self.__myDb.connect()
        self.__db = DbCommand(self.__dbcon, self.__lfh, self.__verbose)

    def reConnect(self):
        """
        Tom : Method to re connect on error with connection
        """
        try:
            self.__myDb.close(self.__dbcon)
        except MySQLdb.Error:
            self.__lfh.write("+WfDbApi.reConnect() DB connection lost - cannot close\n")
            self.__lfh.write("+WfDbApi.reConnect() Re-connecting to the database ..\n")
            self.__lfh.write("+WfDbApi.reConnect() UTC time = %s\n" % datetime.datetime.utcnow())

        for i in range(1, 5):
            try:
                self.__dbcon = self.__myDb.connect()
                self.__db = DbCommand(self.__dbcon, self.__lfh, self.__verbose)
                return True
            except MySQLdb.Error:
                self.__lfh.write("+WfDbApi.reConnect() Cannot get re-connection : trying again\n")
                time.sleep(2 * i)

        return False

    def testConnection(self):
        """
        Test if connection is lost. if yes, rebuild the connection
        """

        if str(self.__dbcon).find("closed") > 0:

            self.__lfh.write("+WfDbApi::testConnection(): Re-connecting to the database.\n")
            self.__dbcon = self.__myDb.connect()
            self.__db = DbCommand(self.__dbcon, self.__lfh, self.__verbose)

    def isConnected(self):
        """Return boolean flag for connection status -"""
        try:
            if str(self.__dbcon).find("open") > 0:
                return True
        except Exception as _e:  # noqa: F841
            pass
        return False

    def close(self):
        """
        This function is only for detecting the connection problem
        """

        if str(self.__dbcon).find("open") > 0:
            self.__myDb.close(self.__dbcon)
            if self.__debug:
                self.__lfh.write("WfDbApi::close(): Closing a connection\n")

    def runInsertSQL(self, sql, args=None):
        """
        method to run a query
        """
        #        self.testConnection()
        for retry in range(1, self.__Nretry):
            ret = self.__db.runInsertSQL(sql, args)
            if ret is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                    # loop again
                else:
                    # unhandled DB error
                    return None
            else:
                return ret

    def runUpdateSQL(self, sql, args=None):
        """
        method to run a query
        """
        # self.testConnection()
        for retry in range(1, self.__Nretry):
            ret = self.__db.runUpdateSQL(sql, args)
            if ret is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                else:
                    # unhandled DB error
                    return None
            else:
                return ret

        # all retries gone bad
        return None

    def runSelectSQL(self, sql):

        """
        method to run a query
        """

        # self.testConnection()
        for retry in range(1, self.__Nretry):
            ret = self.__db.runSelectSQL(sql)
            if ret is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                    # Fall through to loop
                else:
                    # unhandled DB error
                    return None
            else:
                return ret

        # all retries gone bad
        return None

    def getObject(self, depId=None, classId=None, instId=None, taskId=None):
        """
        Get an object by the depId,,classId,instId,taskId
        return a dictionary of table content
        the table is depended on giving depId,classId,instId and taskId

        If all 4 ids are provided, table is wf_task
        if depId,classId,instId provided, but not taskId, table is wf_instance
        if only depId, go to the table deposition
        if classId is provide, depId is empty, the table is wf_class_dict

        Return a dictionary
        The keys are ATTRIBUTES defined in WfSchemaMap._schemaMap and
        The table name is one of (deposition,wf_class_dict,wf_instance,wf_task)

        """

        tableDef = {}
        rDict = {}
        constraintDict = {}
        orderList = []
        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)
        taskId = self.checkId(taskId)

        if depId is not None and classId is not None and instId is not None and taskId is not None:
            # table wf_task
            tableDef = self.__schemaWf[self.__tableList[3]]
        elif depId is not None and classId is not None and instId is not None and taskId is None:
            # table wf_instance
            tableDef = self.__schemaWf[self.__tableList[2]]
        elif depId is None and classId is not None and instId is None and taskId is None:
            # table wf_class_dict (depId must be None)
            tableDef = self.__schemaWf[self.__tableList[1]]
        elif depId is not None and classId is None and instId is None and taskId is None:
            # table deposition
            tableDef = self.__schemaWf[self.__tableList[0]]
        else:
            if self.__verbose:
                self.__lfh.write("WfDbApi::getObject(): Wrong parameters, check all input ids\n")
            return rDict

        if depId is not None:
            constraintDict[self.__idList[0]] = depId
        if classId is not None:
            constraintDict[self.__idList[1]] = classId
        if instId is not None:
            constraintDict[self.__idList[2]] = instId
        if taskId is not None:
            constraintDict[self.__idList[3]] = taskId

        if tableDef == self.__schemaWf[self.__tableList[3]] or tableDef == self.__schemaWf[self.__tableList[2]]:
            orderList.append("ORDINAL")

        # self.testConnection()
        for retry in range(1, self.__Nretry):

            if len(orderList) > 0:
                rDict = self.__db.selectRows(tableDef, constraintDict, orderList)
            else:
                rDict = self.__db.selectRows(tableDef, constraintDict)

            if rDict is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                    # Fall through
                else:
                    # unhandle DB error
                    return None
            else:
                return rDict

        # Retried loop all gone bad
        return None

    def saveObject(self, dataObj, type="insert", constraintDict=None):  # pylint: disable=redefined-builtin
        """
        insert/update a new record in the database
        This method currently is for save one of objects in following list
        deposition, class, instance, task
        """

        for i in [3, 2, 1, 0]:
            if self.__idList[i] in dataObj.keys():
                dataObj[self.__idList[i]] = self.checkId(dataObj[self.__idList[i]])
                if dataObj[self.__idList[i]] is None:
                    self.__lfh.write("WfDbApi::saveObject(): %s can not be None or empty string." % self.__idList[i])
                    exit(1)

        tableDef = self.getTableDef(dataObj)
        if len(tableDef) > 0:

            for retry in range(1, self.__Nretry):

                ret = self.__db.update(type, tableDef, dataObj, constraintDict)

                if ret is None:
                    # database error -
                    if self.__db.dbState > 0:
                        # 2006 : database went away
                        time.sleep(retry * 2)  # backoff for increasing waits
                        if not self.reConnect():
                            return None
                        # else loop
                    else:
                        # unhandled DB error
                        return None
                else:
                    return ret

            # retry limit - all gone bad
            return None

        else:
            self.__lfh.write("WfDbApi::saveObject(): The data object is not the one of deposition, class, instance, task\nNothing is done.\n")
            return "bad-code"

    def getStatus(self, dataObj):
        """
        Get status from a dataObj(could be a deposition, instance or task obj)
        return a string. if dataObj is not a right type, return an empty str.
        """

        returnObj = {}
        oType = str(type(dataObj))
        if oType.find("dict") > 0:
            returnObj = dataObj
        elif oType.find("list") > 0:
            num = len(dataObj)
            # get the last record
            returnObj = dataObj[num - 1]

        for _k, _v in returnObj.items():  # Not clear why for loop
            # a task
            if self.__idList[3] in returnObj.keys() and self.__idList[2] in returnObj.keys() and self.__idList[1] in returnObj.keys() and self.__idList[0] in returnObj.keys():
                return returnObj[self.__statusList[2]]
            # a instance
            elif (
                self.__idList[3] not in returnObj.keys() and self.__idList[2] in returnObj.keys() and self.__idList[1] in returnObj.keys() and self.__idList[0] in returnObj.keys()
            ):
                return returnObj[self.__statusList[1]]
            # a deposition
            elif (
                self.__idList[0] in returnObj.keys()
                and self.__idList[1] not in returnObj.keys()
                and self.__idList[2] not in returnObj.keys()
                and self.__idList[3] not in returnObj.keys()
            ):
                return returnObj[self.__statusList[0]]

            else:
                pass

        return ""

    def updateStatus(self, dataObj, status=None):
        """
        Change status for one entry in table deposition, wf_instance or wf_task

        """
        stype = "update"
        if status is None or status == "":
            self.__lfh.write("WfDbApi::updateStatus(): Failing - no status code given\n")
            exit(1)

        tableDef = self.getTableDef(dataObj)
        if len(tableDef) > 0:
            updateVal = {}
            updateVal = self.assignStatus(tableDef, status)

            rDict = {}
            constraintDict = {}
            if self.__idList[0] in dataObj.keys() and dataObj[self.__idList[0]] is not None and dataObj[self.__idList[0]] != "":
                constraintDict[self.__idList[0]] = dataObj[self.__idList[0]]
            if self.__idList[1] in dataObj.keys() and dataObj[self.__idList[1]] is not None and dataObj[self.__idList[1]] != "":
                constraintDict[self.__idList[1]] = dataObj[self.__idList[1]]
            if self.__idList[2] in dataObj.keys() and dataObj[self.__idList[2]] is not None and dataObj[self.__idList[2]] != "":
                constraintDict[self.__idList[2]] = dataObj[self.__idList[2]]
            if self.__idList[3] in dataObj.keys() and dataObj[self.__idList[3]] is not None and dataObj[self.__idList[3]] != "":
                constraintDict[self.__idList[3]] = dataObj[self.__idList[3]]

            # self.testConnection()
            for retry in range(1, self.__Nretry):

                rDict = self.__db.update(stype, tableDef, updateVal, constraintDict)

                if rDict is None:
                    # database error -
                    if self.__db.dbState > 0:
                        # 2006 : database went away
                        time.sleep(retry * 2)  # backoff for increasing waits
                        if not self.reConnect():
                            return None
                        # else loop
                    else:
                        # unhandle DB error
                        return None
                else:
                    return "ok"

            # retried - all gone bad
            return None

        else:
            self.__lfh.write("+WfDbApi::updateStatus(): The data object is not the one of deposition, instance, task. Nothing is updated.\n")
            return "code-bad"

    def processStatus(self, depID, instID, classID):

        """
        status owner of process
        None : this depID/instID/classID does NOT own the current process
        <value> : the status
        """

        try:
            sql = "select inst_status from dep_last_instance where dep_set_id = '" + str(depID) + "' AND inst_id = '" + str(instID) + "' AND class_id = '" + str(classID) + "'"
            print(sql)
            rows = self.runSelectSQL(sql)
            self.close()
            for row in rows:
                return row[0]

            return None
        except Exception as e:
            print("Exception in processOwner " + str(e))
            return None

    def getReference(self, depId=None, classId=None, instId=None, taskId=None):
        """
        Get a list of reference data from table wf_reference
        An empty id will be treated as None.

        Retrun a rwo list or dictionary (if only one record)
        """

        tableDef = self.__schemaWf[self.__tableList[4]]
        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)
        taskId = self.checkId(taskId)
        constraintDict = {}
        if depId is not None:
            constraintDict[self.__idList[0]] = depId
        else:
            constraintDict[self.__idList[0]] = "None"
        if classId is not None:
            constraintDict[self.__idList[1]] = classId
        else:
            constraintDict[self.__idList[1]] = "None"
        if instId is not None:
            constraintDict[self.__idList[2]] = instId
        else:
            constraintDict[self.__idList[2]] = "None"
        if taskId is not None:
            constraintDict[self.__idList[3]] = taskId
        else:
            constraintDict[self.__idList[3]] = "None"

        # self.testConnection()
        for retry in range(1, self.__Nretry):

            results = self.__db.selectRows(tableDef, constraintDict)

            if results is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                    # loop
                else:
                    # unhandled DB error
                    return None
            else:
                return results

        # retried all gone bad
        return None

    def addReference(self, type, depId=None, classId=None, instId=None, taskId=None, hashId=None, hashVal=None):  # pylint: disable=redefined-builtin
        """
        Add or update reference record in the table wf_reference

        requirement:
        For a deposition level reference,the classId, instId and taskId should be None
        For a class level reference the instId, taskId should be None
        For a instance level reference taskId should be None
        For a task level reference all ids ahould be Non-null.


        """

        tableDef = self.__schemaWf[self.__tableList[4]]
        constraintDict = {}
        updateVal = {}
        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)
        taskId = self.checkId(taskId)

        type = type.lower()
        if type == "update":
            if depId is not None:
                constraintDict[self.__idList[0]] = depId
            else:
                constraintDict[self.__idList[0]] = "None"
            if classId is not None:
                constraintDict[self.__idList[1]] = classId
            else:
                constraintDict[self.__idList[1]] = "None"
            if instId is not None:
                constraintDict[self.__idList[2]] = instId
            else:
                constraintDict[self.__idList[2]] = "None"
            if taskId is not None:
                constraintDict[self.__idList[3]] = taskId
            else:
                constraintDict[self.__idList[3]] = "None"
            if hashId is not None and self.__refList[0] in tableDef["ATTRIBUTES"].keys():
                constraintDict[self.__refList[0]] = hashId
        else:
            if depId is not None:
                updateVal[self.__idList[0]] = depId
            if classId is not None:
                updateVal[self.__idList[1]] = classId
            if instId is not None:
                updateVal[self.__idList[2]] = instId
            if taskId is not None:
                updateVal[self.__idList[3]] = taskId

            if hashId is not None and self.__refList[0] in tableDef["ATTRIBUTES"].keys():
                updateVal[self.__refList[0]] = hashId

        if hashVal is not None and self.__refList[1] in tableDef["ATTRIBUTES"].keys():
            updateVal[self.__refList[1]] = hashVal

        # self.testConnection()
        for retry in range(1, self.__Nretry):

            ret = self.__db.update(type, tableDef, updateVal, constraintDict)

            if ret is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                else:
                    # unhandled DB error
                    return None
            else:
                return "ok"

        # all bad
        return None

    def checkId(self, str):  # pylint: disable=redefined-builtin
        """
        Test if an object id is empty.
        If yes return None, else return str.
        """
        if str == "":
            return None
        else:
            return str

    def exist(self, dataObj):
        """
        Test if a dataObj exists in the Database based on IDs
        dataObj is only for one of objects(deposition, class, instance, task).
        Return Ture or False
        """

        existObj = False
        depId = None
        classId = None
        instId = None
        taskId = None
        if len(dataObj) == 0:
            existObj = False
        else:
            if self.__idList[3] in dataObj.keys():
                taskId = dataObj[self.__idList[3]]
            if self.__idList[2] in dataObj.keys():
                instId = dataObj[self.__idList[2]]
            if self.__idList[1] in dataObj.keys():
                classId = dataObj[self.__idList[1]]
            if self.__idList[0] in dataObj.keys():
                depId = dataObj[self.__idList[0]]
            if depId is None and classId is None and instId is None and taskId is None:
                existObj = False
            else:
                rd = self.getObject(depId, classId, instId, taskId)

            if len(rd) > 0:
                existObj = True
            else:
                existObj = False

        return existObj

    def getTableDef(self, dataObj):
        """
        Determine which table to use in the database

        If all 4 ids are in the dataObj, table is wf_task
        if instId is provided, but not taskId, table is wf_instance
        if only depId, go to the table deposition
        if classId is provide, depId is empty, the table is wf_class_dict
        Return table dict in __schemaWf
        """

        tableDef = {}
        if self.__idList[3] in dataObj.keys() and self.__idList[2] in dataObj.keys() and self.__idList[1] in dataObj.keys() and self.__idList[0] in dataObj.keys():
            tableDef = self.__schemaWf[self.__tableList[3]]
        if self.__idList[3] not in dataObj.keys() and self.__idList[2] in dataObj.keys() and self.__idList[1] in dataObj.keys() and self.__idList[0] in dataObj.keys():
            tableDef = self.__schemaWf[self.__tableList[2]]
        if self.__idList[3] not in dataObj.keys() and self.__idList[2] not in dataObj.keys() and self.__idList[1] in dataObj.keys() and self.__idList[0] not in dataObj.keys():
            tableDef = self.__schemaWf[self.__tableList[1]]
        if self.__idList[3] not in dataObj.keys() and self.__idList[2] not in dataObj.keys() and self.__idList[1] not in dataObj.keys() and self.__idList[0] in dataObj.keys():
            tableDef = self.__schemaWf[self.__tableList[0]]

        return tableDef

    def assignStatus(self, tableDef, status):
        """
        Set status value for table in tableDef

        Return: a dictionary of update Values

        """

        attribDict = tableDef["ATTRIBUTES"]
        updateDict = {}
        for k in range(len(self.__statusList)):
            if self.__statusList[k] in attribDict.keys():
                updateDict[self.__statusList[k]] = status
                # Tom : added change to timestamp when we make a status change
                # Tom : only valid for task and instance : check schema.usefulItems ending in status
                if self.__statusList[k].endswith("STATUS"):
                    updateDict["STATUS_TIMESTAMP"] = getTimeNow()

        return updateDict

    def makeDataDict(self, tableDef, dataObj):
        """
        Make a dictionary to hold update status using table columns as keys

        Return: a dictionary of update values

        """

        attribDict = tableDef["ATTRIBUTES"]
        dataDict = {}
        objType = str(type(dataObj))

        if objType.find("dict") > 0:
            for k, v in dataObj.items():
                if k in attribDict.keys():
                    dataDict[k] = v
        else:
            self.__lfh.write("WfDbApi::makeDataDict(): Failing, input data object is not a dictionary.\n")

        return dataDict

    def getAll(self, depId=None, classId=None, instId=None):
        """
        Get info from WF table deposition, wf_class_dict, wf_instance.

        Return a list of rows (python dictionaries)
        The Keys in the result dictionary is defined in
        WfSchemaMap._selectColumns, could be in the following:
            deposition.dep_set_id
            deposition.pdb_id
            deposition.status_code
            deposition.author_release_status_code
            deposition.exp_method
            deposition.annotator_initials
            wf_class_dict.wf_class_id
            wf_class_dict.wf_class_name
            wf_class_dict.version
            wf_instance.wf_inst_id
            wf_instance.owner
            wf_instance.inst_status
            wf_instance.status_timestamp
            wf_task.wf_task_id
            wf_task.task_name
            wf_task.task_status
            wf_task.status_timestamp
        """

        #
        #   Assumption is that you can't choose classId or instId without
        #   depId. Also any instId must have depId and classId.

        constraintDef = {}

        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)

        if depId is None and (classId is not None or instId is not None):
            self.__lfh.write("+WfDbApi::getAll(): Failing, no deposition id provided\n")
            exit(1)
        if depId is not None and classId is None and instId is not None:
            self.__lfh.write("+WfDbApi::getAll(): Failing, no class id provided\n")
            exit(1)

        if depId is not None:
            constraintDef[self.__idList[0]] = depId
        if classId is not None:
            constraintDef[self.__idList[1]] = classId
        if instId is not None:
            constraintDef[self.__idList[2]] = instId

        # self.testConnection()
        for retry in range(1, self.__Nretry):

            results = self.__db.selectCrossTables(self.__selectList[2], self.__sqlJoinStr, self.__orderBy[2], self.__constraintList, constraintDef)

            if results is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                    # next loop
                else:
                    # unhandled DB error
                    return None
            else:
                return results

        # all gone bad
        return None

    def doQuery(self, level, parameterDict, orderList=None, otherOpt=None):
        """
        Function for determining which table to query based on level.

        Level 1 id for summary information; level 2 is for instance/task
        information.

        parameterDict{} holds parameters that user select from the front end.
        The keys should match the keys in WfSchemaMap._constraintList.

        orderList should also match the keys _schemaMap._orderBy
        or WfSchemaMap._constraintList{} for level 2

        otherOpt could be one of ['AUTHOR_CORRECTIONS','DEP_WITH_PROBLEMS',
        'RELEASE_REQUEST'].

        Return a list of rows (dictionaries)
        The Keys in the result dictionary could be in the following:
        Level 1:
           DEP_SET_ID
           EXP_METHOD
           PDB_ID
           STATUS_CODE
           AUTHOR_RELEASE_STATUS_CODE
           INITIAL_DEPOSITION_DATE
           STATUS_CODE_EXP
           ANNOTATOR_INITIALS
           AUTHOR_LIST
           SG_CENTER
           ASSESSION_CODE
           RELATIONSHIP
           ASSOCIATED_IDS
           CORRECTIONS ( for Author's Corrections)
           REQ_CITATION ( for entries requested for release)
           PROBLEM_TYPE ( for Peorblem/Error entries)
           PROBLEM_DETAILS ( for Peorblem/Error entries)
        Level 2:
           deposition.dep_set_id
           deposition.pdb_id
           deposition.status_code
           deposition.author_release_status_code
           deposition.exp_method
           deposition.annotator_initials
           wf_class_dict.wf_class_id
           wf_class_dict.wf_class_name
           wf_class_dict.version
           wf_instance.wf_inst_id
           wf_instance.owner
           wf_instance.inst_status
           wf_instance.status_timestamp
           wf_task.wf_task_id
           wf_task.task_name
           wf_task.task_status
           wf_task.status_timestamp
        """
        if orderList is None:
            orderList = []

        tableDef = ""
        rList = []
        if level == 2:
            # Multiple tables selected
            # The selectList and query are fixed
            if otherOpt is not None:
                self.__lfh.write("+WfDbApi::doQuery(): Failing, otherOpt is only for level 1\n")
                exit(1)
            for k, _v in parameterDict.items():
                if k not in self.__constraintList.keys():
                    self.__lfh.write("+WfDbApi::doQuery(): Failing, no matched columns in the database for %s\n" % k)
                    exit(1)
            if len(orderList) > 0:
                for k in orderList:
                    if k not in self.__constraintList.keys():
                        self.__lfh.write("+WfDbApi::doQuery(): Failing, no matched columns in the database for %s\n" % k)
                        exit(1)
                orderBy = self.__db.makeOrderStr(orderList)
            else:
                # use default
                orderBy = self.__orderBy[2]

            # self.testConnection()
            ok = False
            for retry in range(1, self.__Nretry):

                rList = self.__db.selectCrossTables(self.__selectList[2], self.__sqlJoinStr, orderBy, self.__constraintList, parameterDict)

                if rList is None:
                    # database error -
                    if self.__db.dbState > 0:
                        # 2006 : database went away
                        time.sleep(retry * 2)  # backoff for increasing waits
                        if not self.reConnect():
                            return None
                        # loop
                    else:
                        # unhandled DB error
                        return None
                else:
                    ok = True
                    break

            # all gone bad
            if not ok:
                return None
        else:
            # level 1, query single tables,
            # first get data from the table "DEPOSITION"
            tableDef = self.__schemaWf[self.__tableList[0]]
            for k in parameterDict.keys():
                if k not in tableDef["ATTRIBUTES"]:
                    self.__lfh.write("+WfDbApi::doQuery(): Failing, no matched columns in the database for %s\n" % (k))
                    exit(1)
            if otherOpt is None:
                if otherOpt in (self.__tableList[9], self.__tableList[10], self.__tableList[11]):
                    sqlStr = self.__idList[0] + " in (select " + self.__idList[0] + " from " + self.__schemaWf[otherOpt]["TABLE_NAME"] + ")"
                    parameterDict["EXTERNAL_TABLE"] = sqlStr
                    #  EXTERNAL_TABLE is for special SQL syntext
                else:
                    self.__lfh.write(
                        "+WfDbApi::doQuery(): Failing, otherOpt should be in the list [%s,%s,%s]\n" % (self.__tableList[9], self.__tableList[10], self.__tableList[11])
                    )
                    exit(1)

            orderList = self.__orderBy[3]
            # self.testConnection()

            ok = False
            for retry in range(1, self.__Nretry):

                results = self.__db.selectRows(tableDef, parameterDict, orderList, self.__selectList[1])

                if results is None:
                    # database error -
                    if self.__db.dbState > 0:
                        # 2006 : database went away
                        time.sleep(retry * 2)  # backoff for increasing waits
                        if not self.reConnect():
                            return None
                    else:
                        # unhandled DB error
                        return None
                else:
                    ok = True
                    break

            # all gone bad
            if not ok:
                return None

            # convert dict to list if there is only one record.
            if str(type(results)).find("dict") > 0:
                rList.append(results)
            else:
                rList = results

            # get info other then the main table

            for k in rList:
                Relations = ""
                associatedIds = ""
                constDict = {}
                DepId = k[self.__selectList[1][0]]
                DepId = self.checkId(DepId)
                if DepId is not None:
                    constDict[self.__idList[0]] = DepId
                    # get 'ASSESSION_CODE' from the table "DATABASE_REF"
                    accessionIds = self.getValueString(self.__schemaWf[self.__tableList[7]], constDict, self.__columnList[3])
                    if accessionIds is not None and accessionIds != "":
                        # fill in 'ASSESSION_CODE' in the result
                        k[self.__columnList[11]] = accessionIds

                    # get 'REPLACE_PDB_ID' from table "DATABASE_PDB_OBS_SPR"
                    obsIds = self.getValueString(self.__schemaWf[self.__tableList[6]], constDict, self.__columnList[4])
                    if obsIds is not None and obsIds != "":
                        Relations += "SPR/OBS"
                        associatedIds += obsIds

                    # get related 'DB_ID' from table "DATABASE_RELATED"
                    # only "SPLIT" is interested
                    constDict[self.__columnList[6]] = "SPLIT"
                    relatedIds = self.getValueString(self.__schemaWf[self.__tableList[8]], constDict, self.__columnList[7])

                    if relatedIds is not None and relatedIds != "":
                        if Relations != "":
                            Relations += ", SPLIT"
                            associatedIds += ", " + relatedIds
                        else:
                            Relations += "SPLIT"
                            associatedIds += relatedIds

                    if Relations is not None and Relations != "":
                        # fill in RELATIONSHIP and ASSOCIATED_IDS in the result
                        k[self.__columnList[9]] = Relations
                        k[self.__columnList[10]] = associatedIds

                    if otherOpt is not None:
                        constDict = {}
                        constDict[self.__idList[0]] = DepId
                        selectItem = ""
                        if otherOpt == self.__tableList[9]:
                            # AUTHOR_CORRECTIONS.CORRECTIONS
                            selectItem = self.__columnList[12]
                        if otherOpt == self.__tableList[10]:
                            # RELEASE_REQUEST.CITATION
                            selectItem = self.__columnList[13]
                        if otherOpt == self.__tableList[11]:
                            # DEP_WITH_PROBLEMS.PROBLEM_TYPE
                            selectItem = self.__columnList[14]

                        resultString = self.getValueString(self.__schemaWf[otherOpt], constDict, selectItem)
                        if resultString is not None and resultString != "":
                            # fill selectItem in the result
                            k[selectItem] = resultString

                        if otherOpt == self.__tableList[11]:
                            # one more items: DEP_WITH_PROBLEMS.PROBLEM_DETAILS
                            selectItem = self.__columnList[15]
                            resultString = self.getValueString(self.__schemaWf[otherOpt], constDict, selectItem)
                            if resultString is not None and resultString != "":
                                k[selectItem] = resultString

        return rList

    def getValueString(self, tableDef, constDict, selectItem):
        """
        Query a single table for a giving DEP_SET_ID
        selectItem is the selected column for query

        return a string (if there are multiple results
        the string is concatenated by ", "
        """

        returnString = ""

        # define the check list 'DATABASE_CODE','REPLACE_PDB_ID','DB_ID'
        # 'CORRECTIONS','REQ_CITATION','PROBLEM_TYPE','PROBLEM_DETAILS',
        # 'PUBMED_ID'
        checkList = [
            self.__columnList[3],
            self.__columnList[4],
            self.__columnList[7],
            self.__columnList[12],
            self.__columnList[13],
            self.__columnList[14],
            self.__columnList[15],
            self.__columnList[16],
        ]
        selectList = []
        if selectItem is not None and selectItem in checkList:
            selectList.append(selectItem)
        else:
            # print "WfDbApi::getValueString(): Warning -- There is not selectItem"
            pass

        # self.testConnection()
        ok = False
        for retry in range(1, self.__Nretry):

            results = self.__db.selectRows(tableDef, constDict, [], selectList)

            if results is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                else:
                    # unhandles DB error
                    return None
            else:
                ok = True
                break

        if not ok:
            return None

        if str(type(results)).find("dict") > 0:
            for checkItem in checkList:
                if checkItem in results.keys():
                    returnString = results[checkItem]
        else:
            for k in results:
                for checkItem in checkList:
                    if checkItem in k.keys():
                        if returnString == "":
                            returnString = k[checkItem]
                        else:
                            returnString += ", " + k[checkItem]

        return returnString

    def getNextWfInstId(self, depId, classId):
        """
        get a WF instance ID in highest number

        return a integer (instId+1)
        """

        rDict = {}
        returnId = 0
        rDict = self.getLastObjectOfState(None, depId, classId)
        if len(rDict) == 0:
            # no instance
            return 1
        else:
            returnId = rDict[self.__idList[2]]
            return int("%s" % returnId[2:]) + 1

    def referenceExist(self, depId=None, classId=None, instId=None, taskId=None, hashId=None, hashVal=None):

        """
        Check to see if the reference data still exists

        """

        tableDef = self.__schemaWf[self.__tableList[4]]
        constraintDict = {}
        selectList = {}
        orderList = {}
        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)
        taskId = self.checkId(taskId)

        if depId is not None:
            constraintDict[self.__idList[0]] = depId
        else:
            constraintDict[self.__idList[0]] = "None"
        if classId is not None:
            constraintDict[self.__idList[1]] = classId
        else:
            constraintDict[self.__idList[1]] = "None"
        if instId is not None:
            constraintDict[self.__idList[2]] = instId
        else:
            constraintDict[self.__idList[2]] = "None"
        if taskId is not None:
            constraintDict[self.__idList[3]] = taskId
        else:
            constraintDict[self.__idList[3]] = "None"

        if hashId is not None and self.__refList[0] in tableDef["ATTRIBUTES"].keys():
            constraintDict[self.__refList[0]] = hashId
        if hashVal is not None and self.__refList[1] in tableDef["ATTRIBUTES"].keys():
            selectList[self.__refList[1]] = hashVal

        # self.testConnection()
        for retry in range(1, self.__Nretry):

            results = self.__db.selectRows(tableDef, constraintDict, orderList, selectList)

            if results is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                else:
                    # unhandle DB error
                    return None
            else:
                if len(results) > 0:
                    return True
                else:
                    return False

        # all gone bad
        return None

    def addReferenceOverwrite(self, depId=None, classId=None, instId=None, taskId=None, hashId=None, hashVal=None):
        """
        This method will insert if the data is new, or update if the data exists

        Add or update reference record in the table wf_reference
        requirement:
        For a deposition level reference,the classId, instId and taskId should be None
        For a class level reference the instId, taskId should be None
        For a instance level reference taskId should be None
        For a task level reference all ids ahould be Non-null.
        """

        refExist = self.referenceExist(depId, classId, instId, taskId, hashId, hashVal)

        if refExist:
            self.addReference("update", depId, classId, instId, taskId, hashId, hashVal)
        else:
            self.addReference("insert", depId, classId, instId, taskId, hashId, hashVal)

    def getLastObjectOfState(self, state, depId, classId, instId=None):
        """
        Get the last instance or task by the depId,,classId and instId
        the table is depended on giving depId,classId,instId

        If instId=None get the last instance from wf_instance;
        if instId != None table is wf_task

        State could be any status in wf_instance or wf_task or None
        Return a dictionary of table content

        """

        tableDef = {}
        rDict = {}
        constraintDict = {}
        orderList = []
        depId = self.checkId(depId)
        if depId is None:
            self.__lfh.write("+WfDbApi::getLastObjectOfState(): WARNING -- No depId provided\n")
            sys.exit(1)

        classId = self.checkId(classId)
        if classId is None:
            self.__lfh.write("+WfDbApi::getLastObjectOfState(): WARNING -- No classId provided\n")
            sys.exit(1)

        if instId is not None:
            instId = self.checkId(instId)

        if depId is not None and classId is not None and instId is not None:
            # table wf_task
            tableDef = self.__schemaWf[self.__tableList[3]]

        elif depId is not None and classId is not None and instId is None:
            # table wf_instance
            tableDef = self.__schemaWf[self.__tableList[2]]
        else:
            self.__lfh.write("+WfDbApi::getLastObjectOfState(): WARNING -- Can not decide the table name\n")
            return rDict

        if depId is not None:
            constraintDict[self.__idList[0]] = depId
        if classId is not None:
            constraintDict[self.__idList[1]] = classId
        if instId is not None:
            constraintDict[self.__idList[2]] = instId
        if state is not None:
            if tableDef == self.__schemaWf[self.__tableList[3]]:
                # wf_task
                constraintDict[self.__statusList[2]] = state

            elif tableDef == self.__schemaWf[self.__tableList[2]]:
                # wf_instance
                constraintDict[self.__statusList[1]] = state
            else:
                self.__lfh.write("+WfDbApi::getLastObjectOfState(): WARNING -- Can not decide the column name about state\n")
                return rDict

        orderList.append("ordinal desc")

        # self.testConnection()
        for retry in range(1, self.__Nretry):

            rDict = self.__db.selectRows(tableDef, constraintDict, orderList)

            if rDict is None:
                # database error -
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry * 2)  # backoff for increasing waits
                    if not self.reConnect():
                        return None
                else:
                    # unhandled DB error
                    return None
            else:
                if str(type(rDict)).find("list") > 0:
                    return rDict[0]
                else:
                    return rDict

        # all gone bad
        return None


if __name__ == "__main__":
    pass



##
# File:  DbApiUtil.py
# Date:  04-May-2015
# Updates:
##
"""
Providing general APIs for database access

This software was developed as part of the World Wide Protein Data Bank
Common Deposition and Annotation System Project

Copyright (c) 2015 wwPDB

This software is provided under a Creative Commons Attribution 3.0 Unported
License described at http://creativecommons.org/licenses/by/3.0/.

"""
__docformat__ = "restructuredtext en"
__author__ = "Zukang Feng"
__email__ = "zfeng@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import os
import sys
import time
import datetime
import MySQLdb

#
from wwpdb.utils.wf.dbapi.DbConnection import DbConnection


class DbApiUtil(object):
    def __init__(self, dbServer=None, dbHost=None, dbName=None, dbUser=None, dbPw=None, dbSocket=None, dbPort=None, verbose=False, log=sys.stderr):
        """ """
        self.__debug = False
        self.__Nretry = 5
        self.__dbServer = dbServer
        self.__dbHost = dbHost
        self.__dbName = dbName
        self.__dbUser = dbUser
        self.__dbPw = dbPw
        self.__dbSocket = dbSocket
        self.__dbPort = dbPort
        self.__verbose = verbose  # pylint: disable=unused-private-member
        self.__lfh = log
        self.__schemaMap = {}
        self.__dbState = 0

        if self.__debug:
            self.__lfh.write("\n+DbApiUtil.__init__() using socket %r\n" % self.__dbSocket)
            self.__lfh.write("+DbApiUtil.__init__() using socket environment reference %r\n" % os.getenv("SITE_DB_SOCKET", None))

        self.__myDb = DbConnection(
            dbServer=self.__dbServer, dbHost=self.__dbHost, dbName=self.__dbName, dbUser=self.__dbUser, dbPw=self.__dbPw, dbPort=self.__dbPort, dbSocket=self.__dbSocket
        )

        self.__dbcon = self.__myDb.connect()

    def __reConnect(self):
        """ """
        try:
            self.__myDb.close(self.__dbcon)
        except MySQLdb.Error:
            self.__lfh.write("+DbApiUtil.reConnect() DB connection lost - cannot close\n")
            self.__lfh.write("+DbApiUtil.reConnect() Re-connecting to the database ..\n")
            self.__lfh.write("+DbApiUtil.reConnect() UTC time = %s\n" % datetime.datetime.utcnow())

        for i in range(1, self.__Nretry):
            try:
                self.__dbcon = self.__myDb.connect()
                self.__dbState = 0
                return True
            except MySQLdb.Error:
                self.__lfh.write("+DbApiUtil.reConnect() Cannot get re-connection : trying again\n")
                time.sleep(2 * i)

        return False

    def __runSelectSQL(self, query):
        """ """
        rows = ()
        try:
            self.__dbcon.commit()
            curs = self.__dbcon.cursor(MySQLdb.cursors.DictCursor)
            curs.execute(query)
            rows = curs.fetchall()
        except MySQLdb.Error as e:
            self.__dbState = e.args[0]
            self.__lfh.write("Database error %s: %s\n" % (e.args[0], e.args[1]))

        return rows

    def __runUpdateSQL(self, query):
        """ """
        try:
            curs = self.__dbcon.cursor()
            curs.execute("set autocommit=0")
            _nrows = curs.execute(query)  # noqa: F841
            self.__dbcon.commit()
            curs.execute("set autocommit=1")
            curs.close()
            return "OK"
        except MySQLdb.Error as e:
            self.__dbcon.rollback()
            self.__dbState = e.args[0]
            self.__lfh.write("Database error %s: %s\n" % (e.args[0], e.args[1]))
        #
        return None

    def setSchemaMap(self, schemaMap):
        """ """
        self.__schemaMap = schemaMap

    def runSelectSQL(self, sql):
        """method to run a query"""
        for retry in range(1, self.__Nretry):
            ret = self.__runSelectSQL(sql)
            if ret is None:
                if self.__dbState > 0:
                    time.sleep(retry * 2)
                    if not self.__reConnect():
                        return None
                else:
                    return None
                #
            else:
                return ret
            #
        #
        return None

    def runUpdateSQL(self, sql):
        """method to run a query"""
        for retry in range(1, self.__Nretry):
            ret = self.__runUpdateSQL(sql)
            if ret is None:
                if self.__dbState > 0:
                    time.sleep(retry * 2)
                    if not self.__reConnect():
                        return None
                else:
                    return None
                #
            else:
                return ret
            #
        #
        return None

    def runUpdate(self, table=None, where=None, data=None):
        if not table:
            return None
        #
        if (not where) and (not data):
            return None
        #
        rowExists = False
        if where:
            sql = "select * from " + str(table) + " where " + " and ".join(["%s = '%s'" % (k, v.replace("'", "\\'")) for k, v in where.items()])
            rows = self.runSelectSQL(sql)
            if rows and len(rows) > 0:
                rowExists = True
            #
        #
        if rowExists and (not data):
            return "OK"
        #
        if rowExists:
            sql = "update " + str(table) + " set " + ",".join(["%s = '%s'" % (k, v.replace("'", "\\'")) for k, v in data.items()])
            if where:
                sql += " where " + " and ".join(["%s = '%s'" % (k, v.replace("'", "\\'")) for k, v in where.items()])
            #
        else:
            sql = "insert into " + str(table) + " (" + ",".join(["%s" % (k) for k, v in where.items()])
            if data:
                sql += "," + ",".join(["%s" % (k) for k, v in data.items()])
            #
            sql += ") values (" + ",".join(["'%s'" % (v.replace("'", "\\'")) for k, v in where.items()])
            if data:
                sql += "," + ",".join(["'%s'" % (v.replace("'", "\\'")) for k, v in data.items()])
            #
            sql += ")"
        #
        return self.runUpdateSQL(sql)

    def runUpdateSQLwithKey(self, key=None, parameter=()):
        """ """
        if not key or not self.__schemaMap or (key not in self.__schemaMap):
            return None
        #
        sql = self.__schemaMap[key]
        if parameter:
            sql = self.__schemaMap[key] % parameter
        #
        return self.runUpdateSQL(sql)

    def selectData(self, key=None, parameter=()):
        """ """
        if key is None or not self.__schemaMap or (key not in self.__schemaMap):
            return None
        #
        sql = self.__schemaMap[key]
        if parameter:
            sql = self.__schemaMap[key] % parameter
        #
        return self.runSelectSQL(sql)



##
# File:    DictUtils.py
# Date:    21-July-2015
#
# Updates:
#
##
"""
Module of extension dictionary translation operations supporting the call protocol of the ProcessRunner() class.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import traceback

from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase
from wwpdb.utils.config.ConfigInfoApp import ConfigInfoAppEm
from mmcif_utils.trans.InstanceMapper import InstanceMapper


class DictUtils(UtilsBase):

    """Utility class to perform extension dictionary content conversions.

    Current supported operations include:
    - cnv-em2emd   PDBx EM dialect to EMD dialect.
    - cnv-emd2em   EMD dialect to PDBx EM dialect

    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=False, log=sys.stderr):
        super(DictUtils, self).__init__(verbose, log)

    def em2emdOp(self, **kwArgs):
        """Convert PDBx EM dialect to EMD dialect."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            #
            dstPath = outObjD["dst"].getFilePathReference()
            #
            cIA = ConfigInfoAppEm()
            mappingInfoPath = cIA.get_emd_mapping_file_path()

            #
            im = InstanceMapper(verbose=self._verbose, log=self._lfh)
            im.setMappingFilePath(mappingInfoPath)
            ok = im.translate(pdbxPath, dstPath, mode="src-dst")
            self._lfh.write("+DictUtils em2emdOp return status %r \n" % ok)
            #
            if self._verbose:
                self._lfh.write("+DictUtils.em2emdOp() - Input model PDBx file path: %s\n" % pdbxPath)
                self._lfh.write("+DictUtils.em2emdOp() - Output result path: %s\n" % dstPath)
            #
            return ok
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def emd2emOp(self, **kwArgs):
        """Convert EMD dialect to PDBx EM dialect."""
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            pdbxPath = inpObjD["src"].getFilePathReference()
            #
            dstPath = outObjD["dst"].getFilePathReference()
            #
            cIA = ConfigInfoAppEm()
            mappingInfoPath = cIA.get_emd_mapping_file_path()

            #
            im = InstanceMapper(verbose=self._verbose, log=self._lfh)
            im.setMappingFilePath(mappingInfoPath)
            ok = im.translate(pdbxPath, dstPath, mode="dst-src")
            self._lfh.write("+DictUtils emd2emOp return status %r \n" % ok)
            #
            if self._verbose:
                self._lfh.write("+DictUtils.emd2emOp() - Input model PDBx file path: %s\n" % pdbxPath)
                self._lfh.write("+DictUtils.emd2emOp() - Output result path: %s\n" % dstPath)
            #
            return ok
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False



"""
   File: DbCommand

   Methods to build and execute relational database queries
   Part of original version written by jdw

   __author__    = "Li Chen"
   __email__     = "lchen@rcsb.rutgers.edu"
   __version__   = "V0.01"
   __Date__      = "April 21, 2010"


Updates :

       16-Dec-2016  jdw - fix string conversion of None values
       26-sep-2017  ep  - runInsertSQL/runUpdateSQL allow for parameterized arguments
                          to pass to db execute to handle quoting
"""

import sys
import MySQLdb


class DbCommand:
    """
    Parameterized SQL queries using Python DBI protocol...
    """

    def __init__(self, dbcon, log=sys.stderr, verbose=False):
        # Tom : put in dbState to store the error code
        self.dbState = 0
        self.__dbcon = dbcon
        self.__lfh = log
        self.__verbose = verbose
        self.__ops = ["EQ", "GE", "GT", "LT", "LE", "LIKE", "NOT LIKE"]
        self.__opDict = {"EQ": "=", "GE": ">=", "GT": ">", "LT": "<", "LE": "<=", "LIKE": "LIKE", "NOT LIKE": "NOT LIKE"}
        self.__logOps = ["AND", "OR", "NOT"]
        self.__grpOps = ["BEGIN", "END"]
        self.__debug = True

    #

    def makeSqlSet(self, attribDict, constraintDef):
        """
        construct syntext in use of SQL commands like 'update' and 'insert'.
        e.g. SET key1 = value1 and key2 = value2 and ...

        Input: attribDict{} defined in WfSchemaMap and constraintDef{}
        Return: a string
        """

        changingVal = ""
        cType = str(type(constraintDef))

        if cType.find("dict") > 0:
            ld = []
            for k, v in constraintDef.items():
                if k in attribDict.keys():
                    if v is None or v == "None":
                        c = " %s = NULL " % (attribDict[k])
                    else:
                        c = " %s = '%s' " % (attribDict[k], v)
                    ld.append(c)
                else:
                    if self.__verbose:
                        self.__lfh.write("DbCommand::makeSqlSet(): Warning -- %s is not defined in the database.\n" % (k))
                if len(ld) > 0:
                    changingVal = " SET " + ld[0]
                    for c in ld[1:]:
                        changingVal += ", " + c

        return changingVal

    def makeConstraintCross(self, constraintList, constraintDef):
        """
        construct constraint syntext in use of SQL commands for
        querying multiple tables
        e.g. WHERE key1 = value1 and key2 > value2 and ...

        Input: constraintList[] defined in WfSchemaMap and constraintDef{}
        Return: a string

        """
        constraint = ""
        ld = []
        for k, v in constraintDef.items():
            if k in constraintList.keys():
                if v is None or v == "None":
                    c = " %s is NULL " % (constraintList[k])
                else:
                    c = " %s = '%s' " % (constraintList[k], v)
                ld.append(c)
            else:
                if self.__verbose:
                    self.__lfh.write("DbCommand::makeConstraintCross(): Warning -- %s is not a key in WfSchemaMap::_constraintList.\n" % (k))

            if len(ld) > 0:
                constraint = " WHERE " + ld[0]
                for c in ld[1:]:
                    constraint += " AND " + c

        return constraint

    def makeSqlConstraint(self, attribDict, constraintDef):
        """
        construct constraint syntext in use for SQL commands.
        e.g. WHERE key1 = value1 and key2 > value2 and ...

        Input: attribDict{} columns defined in WfSchemaMap and
               constraintDef is {} or []
        Return: a string
        """

        constraint = ""
        cType = str(type(constraintDef))

        if cType.find("dict") > 0:
            ld = []
            for k, v in constraintDef.items():
                if k in attribDict.keys():
                    if v == "None" or v is None:
                        c = " %s is NULL " % (attribDict[k])
                    else:
                        c = " %s = '%s' " % (attribDict[k], v)
                    ld.append(c)

                elif k == "EXTERNAL_TABLE":
                    # "EXTERNAL_TABLE" is special designed for SQL syntex
                    # like " column in (select column from another table)"
                    c = "  %s " % (v)
                    ld.append(c)
                else:
                    if self.__verbose:
                        self.__lfh.write("DbCommand::makeSqlConstraint(): Warning -- %s is not defined in the database.\n" % (k))
                if len(ld) > 0:
                    constraint = " WHERE " + ld[0]
                    for c in ld[1:]:
                        constraint += " AND " + c
        elif cType.find("list") > 0:
            #
            # List of tuples with the following syntax:
            #
            # ('EQ'|'GE'|'GT'|'LT'|'LE'|'LIKE', AttribName, Value, type)
            # ('AND'|'OR')
            # ('GROUP', 'BEGIN|END')

            if len(constraintDef) > 0:
                constraint += " WHERE "
                for c in constraintDef:
                    if len(c) == 4 and str(c[0]).upper() in self.__ops:

                        if str(c[3]).upper() == "CHAR":
                            constraint += " %s %s '%s' " % (attribDict[str(c[1]).upper()], self.__opDict[str(c[0]).upper()], str(c[2]))
                        else:
                            constraint += " %s %s %s " % (attribDict[str(c[1]).upper()], self.__opDict[str(c[0]).upper()], str(c[2]))

                    elif len(c) == 3 and str(c[0]).upper() in self.__ops:
                        constraint += " %s %s '%s' " % (attribDict[str(c[1]).upper()], self.__opDict[str(c[0]).upper()], str(c[2]))
                    elif len(c) == 2 and str(c[0]).upper() == "GROUP" and str(c[1]).upper() in self.__grpOps:  # noqa: W504
                        if str(c[1]).upper() == "BEGIN":
                            constraint += "("
                        else:
                            constraint += ")"
                    elif len(c) == 2 and str(c[0]).upper() == "LOGOP" and str(c[1]).upper() in self.__logOps:  # noqa: W504
                        constraint += " %s " % str(c[1]).upper()
                    else:
                        if self.__lfh:
                            self.__lfh.write("Constraint error: %s\n" % str(c))

        else:
            #           Just ignore if contraints are entered as None
            #           if(self.__verbose):
            self.__lfh.write("DbCommand::makeSqlConstraint(): Warning -- constraint type error: %s\n" % str(cType))

        return constraint

    def makeOrderStr(self, orderList):
        """
        construct a string "ORDER BY ..." in SQL command for cross tables search

        """
        orderBy = ""
        if len(orderList) > 0:
            orderBy = " ORDER BY " + orderList[0]
            for c in orderList[1:]:
                orderBy += ", " + c

        return orderBy

    def runInsertSQL(self, query, args=None):
        return self.runUpdateSQL(query, args)

    def runUpdateSQL(self, query, args=None):
        try:
            curs = self.__dbcon.cursor()
            curs.execute("set autocommit=0")
            if args:
                nrows = curs.execute(query, args)
            else:
                nrows = curs.execute(query)
            self.__dbcon.commit()
            curs.execute("set autocommit=1")
            curs.close()
            return nrows
        except MySQLdb.Error as e:
            self.__dbcon.rollback()
            # TOm : no curs defined
            #            curs.execute("set autocommit=1")
            #            curs.close()
            self.dbState = e.args[0]
            return None

    def runSelectSQL(self, query):
        """
        method to run a SQL query : no checking - just a convience method to
        get things working
        note that this does not use attrib list = but returns a list of lists.
        """
        returnList = []

        if self.__debug:
            self.__lfh.write("\n+DbCommand.runSelectSQL - input query %r\n" % query)
        if query is not None:
            row = []
            try:
                self.__dbcon.commit()
                curs = self.__dbcon.cursor()
                curs.execute(query)
                while True:
                    result = curs.fetchone()
                    if result is not None:
                        row = []
                        ir = 0
                        for _k in result:
                            row.append(result[ir])
                            ir += 1
                        returnList.append(row)
                    else:
                        break
            except MySQLdb.Error as e:
                self.__lfh.write("DbCommand::runSelectSQL(): Database error %s: %s\n" % (e.args[0], e.args[1]))
                self.__lfh.write("DbCommand::runSelectSQL(): Failing on query %s\n" % query)
                # no curs defined here
                #                curs.close()
                self.__dbcon.close()
                self.dbState = e.args[0]
                return None
        #            sys.exit (1)
        #
        if self.__debug:
            self.__lfh.write("+DbCommand.runSelectSQL - result length is %s\n" % len(returnList))

        return returnList

    def selectRows(self, tableDef, constraintDef, orderList=None, selectList=None):
        """
        Execute query on the table described by "tableDef"
        subject to the conditions in constraintDef and
        sorted by the list of attributes in orderList [(attrib,type),..].
        An optional selectList=[attrib,attrib] can be provided to limit
        the query selection.

        ConstraintDef can be a simple of dictionary of key == value pairs
        which are logically AND'd together, or a more general constraint
        can be contructed using the compact contraint specification list
        decoded in method makeSqlConstraint() above.

        Return a <row list or row dictionary>.
        """
        if orderList is None:
            orderList = []
        if selectList is None:
            selectList = []

        tableName = tableDef["TABLE_NAME"]
        attribDict = tableDef["ATTRIBUTES"]
        if len(selectList) > 0:
            attribsCsv = ",".join(["%s" % attribDict[k] for k in selectList])
            attribs = selectList
        else:
            attribsCsv = ",".join(["%s" % v for v in attribDict.values()])
            attribs = attribDict.keys()

        #
        #  Build selection constraints ...
        #
        constraint = self.makeSqlConstraint(attribDict, constraintDef)

        order = ""
        if len(orderList) > 0:
            order = " ORDER BY " + orderList[0]
            for a in orderList[1:]:
                order += ", " + a
            # order += " desc "

        #
        query = "SELECT " + attribsCsv + " FROM " + tableName + constraint + order
        # Tom added verbose check
        if self.__verbose:
            self.__lfh.write("DB command --\n%s\n" % query)

        ##
        returnList = []
        row = {}
        try:
            self.__dbcon.commit()
            curs = self.__dbcon.cursor()
            curs.execute(query)
            while True:
                result = curs.fetchone()
                if result is not None:
                    row = {}
                    ir = 0
                    for k in attribs:
                        row[k] = result[ir]
                        ir += 1
                    returnList.append(row)
                else:
                    break
        except MySQLdb.Error as e:
            self.__lfh.write("DbCommand::selectRows(): Database error %s: %s\n" % (e.args[0], e.args[1]))
            self.__lfh.write("DbCommand::selectRows(): Failing on query %s\n" % query)
            # Tom : no curs defined here
            #            curs.close()
            self.__dbcon.close()
            self.dbState = e.args[0]
            return None
        #            sys.exit (1)

        if len(returnList) > 1:
            return returnList
        else:
            return row

    def update(self, type, tableDef, updateVal, constraintDef=None):  # pylint: disable=redefined-builtin
        """
        Update value for any column(s) in a giving table.
        type may be 'insert' or 'update', default is 'insert'.
        tableDef in WfSchemaMap, updateVal{} and constraintDef{}
        for "update"
        """

        tableName = tableDef["TABLE_NAME"]
        attribDict = tableDef["ATTRIBUTES"]
        # modify the specicial characters like "'",""" in the updateVal
        updateValMod = {}
        v2 = ""
        for k, v in updateVal.items():
            if str(v).find("'") != -1:
                v2 = str(v).replace("'", "\\'")
            elif str(v).find('"') != -1:
                v2 = v.replace('"', '\\"')
            elif v is not None:
                v2 = str(v)
            else:
                v2 = v
            updateValMod[k] = v2

        updateSet = self.makeSqlSet(attribDict, updateValMod)
        constraint = ""
        if constraintDef is not None:
            constraint = self.makeSqlConstraint(attribDict, constraintDef)
        if type.lower() == "update":
            command = "UPDATE "
        else:
            command = "INSERT INTO "
        command += tableName + updateSet
        if constraint != "":
            command += constraint
        # Tom - added verbose check
        if self.__verbose:
            self.__lfh.write("DB command --\n%s\n" % command)

        try:

            curs = self.__dbcon.cursor()
            curs.execute("set autocommit=0")
            curs.execute(command)

        except MySQLdb.Error as e:
            self.__lfh.write("DbCommand::update(): Database error %s: %s\n" % (e.args[0], e.args[1]))

            self.__dbcon.rollback()
            # no curs defined here
            #            curs.execute("set autocommit=1")
            #            curs.close()
            self.__dbcon.close()
            self.dbState = e.args[0]
            return None

        self.__dbcon.commit()
        curs.execute("set autocommit=1")
        if self.__verbose:
            self.__lfh.write("DbCommand::update(): SQL command successfully executed.\n")
        curs.close()

        return "ok"

    def selectCrossTables(self, selectList, sqlJoinStr, orderBy, constraintList, constraintDef=None):
        """
        This function is specically for some complicate queries. The
        "sql join" is used.

        Input selectList, sqlJoinStr, orderBy will be
        specially definded in WfSchemaMap. constraintDict is {}

        Return a list of rows {dictionaries)

        """

        attribsCsv = ",".join(["%s" % k for k in selectList])

        constraint = self.makeConstraintCross(constraintList, constraintDef)
        query = "SELECT DISTINCT " + attribsCsv + sqlJoinStr + constraint + orderBy
        # Tom - added verbose check
        if self.__verbose:
            self.__lfh.write("DB command --\n%s\n" % query)

        returnList = []
        row = {}

        try:
            self.__dbcon.commit()
            curs = self.__dbcon.cursor()
            curs.execute(query)
            while True:
                result = curs.fetchone()
                if result is not None:
                    row = {}
                    ir = 0
                    for k in selectList:
                        row[k] = result[ir]
                        ir += 1
                    returnList.append(row)
                else:
                    break
        except MySQLdb.Error as e:
            self.__lfh.write("DbCommand::selectRows(): Database error %s: %s\n" % (e.args[0], e.args[1]))
            # Tom : no curs defined here
            #          curs.close()
            self.dbState = e.args[0]
            return None

        return returnList



#
# File:   SFConvert.py
# Date:   2021-03-01
#
"""
Module to handle dictionary compliant SF files from autoPROC in mmCIF format
"""


import os
import sys

try:
    # We will have present on annotation system - but allow testing of DepUI merge without
    from wwpdb.apps.ann_tasks_v2.expIoUtils.PdbxExpIoUtils import PdbxExpFileIo, PdbxExpIoUtils
except ImportError:
    pass
from mmcif.io.IoAdapterCore import IoAdapterCore


class SFConvert(object):
    def __init__(self, verbose=False, log=sys.stderr):
        self._verbose = verbose
        self._lfh = log

    def isSpecialSF(self, pathIn):
        """Detect if autoPROC sf file uploaded

        Returns True if special handling warranted
        """
        try:
            io = IoAdapterCore()
            # Categories that will trigger special handling.  AUTOProc uses software
            # Used to key off software
            catList = ["pdbx_audit_conform"]

            containerList = io.readFile(pathIn, selectList=catList)

            if containerList is None or len(containerList) < 1:
                return False

            for cat in catList:
                for cl in containerList:
                    if cat in cl.getObjNameList():
                        return True
        except Exception as _e:  # noqa: F841
            pass

        return False

    def handleSpecialSF(self, sfIn, sfOut, modelIn, sfDiagFilePath, logFilePath):

        # Empty log and diagnostic
        with open(sfDiagFilePath, "w") as _fout:  # noqa: F841
            pass
        with open(logFilePath, "w") as _fout:  # noqa: F841
            pass

        # Get pdbid
        pdbId = self.__getmodelPdbId(modelIn)
        if pdbId is None:
            pdbId = "xxxx"

        try:
            sfIo = PdbxExpFileIo(verbose=self._verbose, log=self._lfh)
            containerList = sfIo.getContainerList(sfIn)
            if len(containerList) < 1:
                return False

            sfIo.updateContainerNames(idCode=pdbId, containerList=containerList)
            sfIo.updateEntryIds(idCode=pdbId, containerList=containerList)
            self.__stripAudit(containerList)
            self.__correctGphlAlias(containerList)
            _err = self.__checkErrors(containerList, sfDiagFilePath, logFilePath)  # noqa: F841
            # We always write to avoid complaint about pdbx2pdbx failing
            sfIo.writeContainerList(sfOut, containerList)
            return True
        except Exception as e:
            self._lfh.write("Failed to convert %s\n" % e)

        return False

    def __getmodelPdbId(self, modelIn):
        """Gets pdb id if present"""
        try:
            if modelIn and os.path.exists(modelIn):
                mIo = PdbxExpFileIo(verbose=self._verbose, log=self._lfh)
                modelContainerList = mIo.getContainerList(modelIn)
                if len(modelContainerList) < 1:
                    return None

                mE = PdbxExpIoUtils(dataContainer=modelContainerList[0], verbose=self._verbose, log=self._lfh)
                pdbId = str(mE.getDbCode(dbId="PDB")).lower()

                return pdbId
        except Exception as _e:  # noqa: F841
            pass

        return None

    def __stripAudit(self, cList):
        """Strip out Audit records from uploaded files"""
        for cat in ["audit", "audit_conform"]:
            for container in cList:
                if cat in container.getObjNameList():
                    container.remove(cat)

    def __correctGphlAlias(self, cList):
        """Strip out Audit records from uploaded files"""

        rmappings = {"gphl_signal_type": "pdbx_signal_type", "gphl_observed_signal_threshold": "pdbx_observed_signal_threshold"}

        for container in cList:
            if "reflns" in container.getObjNameList():
                cObj = container.getObj("reflns")
                cObj.renameAttributes(rmappings)

    def __checkErrors(self, cList, sfDiagFilePath, logFilePath):
        """Checks for standard errors in files"""

        errs = []

        if len(cList) < 1:
            errs.append("Error: empty file")

        else:
            # In first datablock, refln category should have at least one of: F/I/F+/F-/I+/I-
            b0 = cList[0]

            if "refln" not in b0.getObjNameList():
                errs.append("Error: missing _refln category in first block")
            else:
                cObj = b0.getObj("refln")
                needAtt = ["F_meas_au", "F_meas", "intensity_meas", "intensity_meas_au", "F_squared_meas", "pdbx_I_plus", "pdbx_I_minus", "pdbx_F_plus", "pdbx_F_minus"]
                found = False
                for att in needAtt:
                    if att in cObj.getAttributeList():
                        found = True
                if not found:
                    errs.append("Error: SF file missing mandatory items 'F/I/F+/F-/I+/I-'")

            # Check all datablocks to ensure none have both merged and unmerged data
            for bl in cList:
                names = bl.getObjNameList()
                if "refln" in names and "diffrn_refln" in names:
                    errs.append("Error: cannot have _refln and _diffrn_refln categories in same data block")

        if len(errs) > 0:
            # sfDiagFilePath, logFilePath):
            with open(sfDiagFilePath, "a") as fOut:
                for e in errs:
                    fOut.write("%s\n" % e)

            with open(logFilePath, "a") as fOut:
                for e in errs:
                    fOut.write("%s\n" % e)

            return True

        return False



"""
      File: DbConnection

   Database connection class (MYSQL only)

   __author__    = "Li Chen"
   __email__     = "lchen@rcsb.rutgers.edu"
   __version__   = "V0.01"
   __Date__      = "April 21, 2010"

  Updated:

     07-Feb-2014  jdw  -  Add socket support --
     23-Mar-2016  jdw  -  make ports ints.
"""
import sys
import os
import MySQLdb


class DbConnection:
    """Class to encapsulate rdbms DBI connection ..."""

    def __init__(self, dbServer="mysql", dbHost="localhost", dbName=None, dbUser=None, dbPw=None, dbPort=None, dbSocket=None, log=sys.stderr, verbose=True):
        self.__lfh = log
        self.__verbose = verbose  # pylint: disable=unused-private-member
        self.__debug = False
        if self.__debug:
            self.__lfh.write(
                "\n\n+DbConnection.connect(): Connection using server %s host %s dsn %s user %s pw %s port %d socket %s\n"
                % (dbServer, dbHost, dbName, dbUser, dbPw, dbPort, dbSocket)
            )

        if dbName is None:
            self.__dbName = os.getenv("MYSQL_DB_NAME")
        else:
            self.__dbName = dbName

        if dbUser is None:
            self.__dbUser = os.getenv("MYSQL_DB_USER")
        else:
            self.__dbUser = dbUser

        if dbPw is None:
            self.__dbPw = os.getenv("MYSQL_DB_PW")
        else:
            self.__dbPw = dbPw

        if dbHost is None:
            self.__dbHost = os.getenv("MYSQL_DB_HOST")
        else:
            self.__dbHost = dbHost

        if dbPort is None:
            self.__dbPort = os.getenv("MYSQL_DB_PORT")
        else:
            self.__dbPort = dbPort

        try:
            self.__dbPort = int(self.__dbPort)
        except Exception as _e:  # noqa: F841
            pass

        if dbSocket is None:
            # try from the environment -
            tS = os.getenv("MYSQL_DB_SOCKET")
            if self.__debug:
                self.__lfh.write("+DbConnection.__init__ MYSQL_DB_SOCKET is %r\n" % tS)
            if tS is not None:
                self.__dbSocket = tS
            else:
                self.__dbSocket = None
        else:
            self.__dbSocket = dbSocket

        self.__dbServer = dbServer

        if dbServer != "mysql":
            self.__lfh.write("DbConnection::__init__(): unsupported server %s\n" % dbServer)
            sys.exit(1)

        self.__dbcon = None

    def connect(self):
        """Consistent db connection method...

        Return a connection object
        """
        try:
            if self.__dbSocket is None:
                dbcon = MySQLdb.connect(
                    db="%s" % self.__dbName, user="%s" % self.__dbUser, passwd="%s" % self.__dbPw, port=self.__dbPort, host="%s" % self.__dbHost, local_infile=1
                )
            else:
                dbcon = MySQLdb.connect(
                    db="%s" % self.__dbName,
                    user="%s" % self.__dbUser,
                    passwd="%s" % self.__dbPw,
                    port=self.__dbPort,
                    host="%s" % self.__dbHost,
                    unix_socket="%s" % self.__dbSocket,
                    local_infile=1,
                )

        except MySQLdb.Error as e:
            self.__lfh.write("+DbConnection.connect(): Connection error %s: %s\n" % (e.args[0], e.args[1]))
            self.__lfh.write(
                "+DbConnection.connect(): Connection failed using server %s host %s dsn %s user %s pw %s port %d socket %s\n"
                % (self.__dbServer, self.__dbHost, self.__dbName, self.__dbUser, self.__dbPw, self.__dbPort, self.__dbSocket)
            )
            sys.exit(1)

        return dbcon

    def close(self, dbcon):
        """
        Consistent db close connection method...
        """
        dbcon.close()

    def commit(self):
        """
        Consistent db commit() method...
        """
        self.__dbcon.commit()

    def rollback(self):
        """
        Consistent db rollback() method...
        """
        self.__dbcon.rollback()



##
# File:  dbApi.py
# Date:  14-Feb-2015
#
# #

import sys
import logging

from wwpdb.utils.wf.dbapi.WfDbApi import WfDbApi

logger = logging.getLogger(__name__)
"""

Grab bag of methods to execute SQL commands on various WF status tables -

"""


class dbAPI(object):
    def __init__(self, depID, connection=None, verbose=True):

        if connection:
            self.con = connection
        else:
            self.con = WfDbApi(verbose=False)
        self.depID = depID
        self.verbose = verbose

    def close(self):

        self.con.close()

    def runSelectNQ(self, table=None, join=None, select=None, where=None, order=None, reverse=False, ordinal=0, run=True, limit=0):

        if where:
            for k, v in where.items():
                where[k] = "'" + v + "'"
        return self.runSelect(table=table, join=join, select=select, where=where, order=order, reverse=reverse, ordinal=ordinal, run=run, limit=limit)

    def runSelect(self, table=None, join=None, select=None, where=None, order=None, reverse=False, ordinal=0, run=True, limit=0):
        """
        Very simple select SQL creator - single table
        table = string
        values = list
        where = dictionary of equivalence
        order = list

        Joins are specified with multiple table string , and where clause with inner join defined

        IMPT : all attribute values MUST BE QUOTED if strings

        Returns a list of list  (rows of data columns)
        Returns an empty list if nothing returned
        """

        if not table:
            logger.info("WFE.dbAPI.runSelect : Undefined table")
            return []

        if not select:
            logger.info("WFE.dbAPI.runSelect : Undefined select")
            return []

        depDB = {}
        depDB["DEP_SET_ID"] = self.depID

        try:
            if self.con.exist(depDB):
                #       if True:
                sql = "select " + ",".join(select) + " from " + str(table) + " "
                if where:
                    sql += " where " + " and ".join(["%s = %s" % (k, v) for k, v in where.items()])
                if join:
                    sql += " and " + join
                if ordinal > 0:
                    sql += " and ordinal >= " + str(ordinal)
                if order:
                    sql += " order by " + ",".join(order)
                    if reverse:
                        sql += " desc "
                if limit > 0:
                    sql += " limit " + str(limit)
                if self.verbose:
                    logger.info("WFE.dbAPI.runSelect > %s", str(sql))
                if run:
                    ret = self.con.runSelectSQL(sql)
                    return ret
                else:
                    return sql
            else:
                return []
        except Exception as e:
            logger.exception("WFE.dbAPI.runSelect :Exception %s", str(e))
            return []

    def runUpdateOnOrdinal(self, table=None, ordinal=None, data=None, run=True):
        """
        Very simple update/insert SQL creator base on ordinal
        table = string
        data = values to insert/update
        where clause is hard wired to ordinal (unique) so can only UPDATE
            a bad ordinal returns 0 rows udpated

        return number of rows updated
        if run = False : returns the SQL

        IMPT : all attribute values MUST BE QUOTED if strings
        """

        if not table:
            logger.info("WFE.dbAPI.runUpdateOnOrdinal : Undefined table")
            return False

        if not ordinal:
            logger.info("WFE.dbAPI.runUpdateOnOrdinal : Undefined ordinal")
            return False

        try:
            sql = "update " + str(table) + " set " + ",".join(["%s = %s" % (k, v) for k, v in data.items()])
            sql = sql + " where ordinal = " + str(ordinal)
            if self.verbose:
                logger.info("WFE.dbAPI.runInsertUpdate(update) > %s", str(sql))

            if run:
                ok = self.con.runUpdateSQL(sql)
                if not ok:
                    logger.info("WFE.dbAPI.runSelect :False to update/insert data %s", str(sql))
                return ok
            else:
                return sql
        except Exception as e:
            logger.info("WFE.dbAPI.runSelect :Exception %s", str(e))
            return False

    def runInsertUpdateNQ(self, table=None, depID=None, where=None, data=None, run=True):
        """
        wrapper to add quotes to values
        """

        if where:
            for k, v in where.items():
                where[k] = "'" + v + "'"
        if data:
            for k, v in data.items():
                data[k] = "'" + v + "'"
        return self.runInsertUpdate(table, depID, where, data, run)

    def runInsertNQ(self, table=None, depID=None, where=None, data=None, run=True):

        if where:
            for k, v in where.items():
                where[k] = "'" + v + "'"
        if data:
            for k, v in data.items():
                data[k] = "'" + v + "'"
        return self.runInsert(table, depID, where, data, run)

    def runInsert(self, table=None, depID=None, where=None, data=None, run=True):

        try:
            if depID:
                sql = "insert into " + str(table) + " (dep_set_id," + ",".join(["%s" % (k) for k, v in data.items()])
                sql += ") values ('" + str(depID) + "'," + ",".join(["%s" % (v) for k, v in data.items()]) + ")"

            if where:
                sql = "insert into " + str(table) + " (" + ",".join(["%s" % (k) for k, v in where.items()])
                if data:
                    sql += "," + ",".join(["%s" % (k) for k, v in data.items()])
                sql += ") values (" + ",".join(["%s" % (v) for k, v in where.items()])
                if data:
                    sql += "," + ",".join(["%s" % (v) for k, v in data.items()])
                sql += ")"

            if self.verbose:
                logger.info("WFE.dbAPI.runInsertUpdate(insert) > %s", str(sql))

            if run:
                return self.con.runInsertSQL(sql)
            else:
                return sql
        except Exception as e:
            logger.exception("WFE.dbAPI.runInsert :Exception %s", str(e))
            return False

    def runUpdate(self, table=None, depID=None, where=None, data=None, run=True):
        logger.debug("Beginning run update")
        try:
            sql = "update " + str(table) + " set " + ",".join(["%s = %s" % (k, v) for k, v in data.items()])
            if depID:
                sql += " where dep_set_id = '" + str(depID) + "'"
            if where:
                sql += " where " + " and ".join(["%s = %s" % (k, v) for k, v in where.items()])

            if self.verbose:
                logger.info("WFE.dbAPI.runInsertUpdate(update) > %s", str(sql))

            if run:
                return self.con.runUpdateSQL(sql)
            else:
                return sql

        except Exception as e:
            logger.exception("WFE.dbAPI.runUpdate :Exception %s", str(e))
            return False

    def runInsertUpdate(self, table=None, depID=None, where=None, data=None, run=True):
        """
        Very simple update/insert SQL creator
        table = string
        data = values to insert/update
        where clause is hard wired to dep_set_id so can ONLy be used for
            for tables with unique rows on dep_set_id
            Ie if the table contains a row with depsetid - then it is update
                  if not - the data is inserted

        return number of rows updated
        if run = False : returns the SQL

        if depID is not none - then we test the existance based on depID and the table MUST BE UNIQUE on depID
          data is added to the row based on the dep_set_id + data
        if unique is not none - then we test the existance on the synthetic key based on the unique dictionary
          data is added to the row based on the where + data


        IMPT : all attribute values MUST BE QUOTED if strings
        """

        if not table:
            logger.info("WFE.dbAPI.runUpdateOnOrdinal : Undefined table")
            return False

        rowExists = False
        if depID:
            # then data is unique on depID
            depDB = {}
            depDB["DEP_SET_ID"] = depID
            if self.con.exist(depDB):
                rowExists = True

        else:
            if where:
                sql = "select ordinal from " + str(table) + " where " + " and ".join(["%s = %s" % (k, v) for k, v in where.items()])
                rows = self.con.runSelectSQL(sql)
                if rows and len(rows) > 0:
                    rowExists = True
            else:
                logger.info("WFE.dbAPI.runUpdate: Undefined key ")

        ok = True
        try:
            if rowExists:
                ok = self.runUpdate(table, depID, where, data, run)

            else:
                ok = self.runInsert(table, depID, where, data, run)

            if not run:
                return ok

            if not ok:
                logger.info("WFE.dbAPI.runSelect :False to update/insert data ")
                return ok
            else:
                return ok
        except Exception as e:
            logger.exception("WFE.dbAPI.runSelect :Exception %s", str(e))
            return False


def main(_argv):

    print("starting DBAPI test")

    depid = "D_1100201819"
    ss = dbAPI(depid, verbose=True)

    # table join test - notice that in the where clause be careful with the key must not repeat - so be carful with the inner join
    ret = ss.runSelect(
        table="deposition,wf_instance",
        select=["deposition.dep_set_id", "wf_instance.wf_inst_id"],
        where={"deposition.dep_set_id": "wf_instance.dep_set_id", "wf_instance.dep_set_id": "'" + depid + "'", "wf_instance.wf_inst_id": "'W_001'"},
        run=True,
    )
    print(str(ret))

    ret = ss.runSelectNQ(table="deposition", select=["ordinal", "dep_set_id", "depPW", "annotator_initials"], where={"dep_set_id": depid})
    print(str(ret))

    # test the orinal updates
    ret = ss.runSelectNQ(table="deposition", select=["ordinal", "dep_set_id", "depPW"], where={"dep_set_id": depid})
    print(str(ret))
    #   ret = ss.runUpdateOnOrdinal(table='deposition',ordinal='17823',data={"depPW":"'abcdef'"},run=True)
    #   print str(ret)
    ret = ss.runSelectNQ(table="deposition", select=["ordinal", "dep_set_id", "depPW"], where={"dep_set_id": depid})
    print(str(ret))
    #   ret = ss.runUpdateOnOrdinal(table='deposition',ordinal='17823',data={"depPW":"'123456'"},run=True)
    print(str(ret))
    ret = ss.runSelectNQ(table="deposition", select=["ordinal", "dep_set_id", "depPW"], where={"dep_set_id": depid})
    print(str(ret))

    ret = ss.runSelectNQ(
        table="deposition,user_data", select=["last_name", "role", "user_data.country"], join="deposition.dep_set_id=user_data.dep_set_id", where={"deposition.dep_set_id": depid}
    )

    # test the dep_set_id updates
    #   ret = ss.runInsertUpdate(table='deposition',depID='D_1000200025',data={"depPW":"'abcdef'"},run=True)
    print(str(ret))
    ret = ss.runSelect(table="deposition", select=["dep_set_id", "depPW"], where={"dep_set_id": "'" + depid + "'"})
    print(str(ret))
    #   ret = ss.runInsertUpdate(table='deposition',depID='D_1000200025',data={"depPW":"'123456'"},run=True)
    print(str(ret))
    ret = ss.runSelect(table="deposition", select=["dep_set_id", "depPW"], where={"dep_set_id": "'" + depid + "'"})
    print(str(ret))

    print("finished")


if __name__ == "__main__":
    main(sys.argv[1:])



##
# File:    EmUtils.py
# Date:    10-Feb-2014
#
# Updates:
#
#    15-Aug-2014 JDW add mapFixInPlaceOp(self,**kwArgs)
#    31-Aug-2015 jdw add mapFixInPlaceAltOp()
#    02-Oct-2015 jdw add mapFixInPlaceCfgOp
#    07-Oct-2015 jdw change api call in mapFixInPlaceCfgOp to deposit-update-map-header-in-place
##
"""
Module of EM utility operations supporting the call protocol of the ProcessRunner() class.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import json
import traceback

from wwpdb.utils.wf.plugins.UtilsBase import UtilsBase
from wwpdb.utils.config.ConfigInfo import ConfigInfo

from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility
from wwpdb.io.locator.PathInfo import PathInfo


class EmUtils(UtilsBase):

    """Utility class to run validation operations.

    Current supported operations include:

    - create PDF report and supporting XML data file

    Each method in this class implements the method calling interface of the
    `ProcessRunner()` class.   This interface provides the keyword arguments:

    - inputObjectD   dictionary of input objects
    - outputObjectD  dictionary of output objects
    - userParameterD  dictionary of user adjustable parameters
    - internalParameterD dictionary of internal parameters

    Each method in the class handles its own exceptions and returns
    True on success or False otherwise.

    """

    def __init__(self, verbose=True, log=sys.stderr):
        super(EmUtils, self).__init__(verbose, log)
        self.__cleanUp = False
        """Flag to remove any temporary directories created by this class.
        """

    def fsc_check_options(self, **kwArgs):
        """
        Set the options for xmllint
        """

        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            inputPath = inpObjD["src"].getFilePathReference()
            outPath = outObjD["dst1"].getFilePathReference()
            dirPath = outObjD["dst1"].getDirPathReference()
            logPath = outObjD["dst2"].getFilePathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(inputPath)
            #  add any extra command line options
            options = str(uD["options"])
            if (options is not None) and (len(options) > 0):
                dp.addInput(name="options", value=options)

            ret = dp.op("fsc_check")
            dp.expLog(logPath)
            dp.exp(outPath)

            if self._verbose:
                self._lfh.write("+EmUtils.fsc_check_options() - input FSC file path:  %s\n" % inputPath)
                self._lfh.write("+EmUtils.fsc_check_options() - output FSC file path:  %s\n" % outPath)
                self._lfh.write("+EmUtils.fsc_check_options() - log file path:         %s\n" % logPath)

            if self.__cleanUp:
                dp.cleanup()

            return ret
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return -100

    def mapFixOp(self, **kwArgs):
        """Run mapfix (Prior BIG VERSION) to correct header values in input map."""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            inpMapPath = inpObjD["src"].getFilePathReference()
            outMapPath = outObjD["dst1"].getFilePathReference()
            dirPath = outObjD["dst1"].getDirPathReference()
            logPath = outObjD["dst2"].getFilePathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(inpMapPath)
            #  add any extra command line options
            options = str(uD["options"])
            if (options is not None) and (len(options) > 0):
                dp.addInput(name="options", value=options)

            dp.op("mapfix-big")
            dp.expLog(logPath)
            dp.exp(outMapPath)

            if self._verbose:
                self._lfh.write("+EmUtils.mapFixOp() - input map  file path:  %s\n" % inpMapPath)
                self._lfh.write("+EmUtils.mapFixOp() - output map file path:  %s\n" % outMapPath)
                self._lfh.write("+EmUtils.mapFixOp() - log file path:         %s\n" % logPath)

            if self.__cleanUp:
                dp.cleanup()
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def mapFixInPlaceOp(self, **kwArgs):
        """Run mapfix (BIG VERSION) to correct header values in input map (in place)."""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            inpMapPath = inpObjD["src"].getFilePathReference()
            outMapPath = outObjD["dst1"].getFilePathReference()
            dirPath = outObjD["dst1"].getDirPathReference()
            rptPath = outObjD["dst2"].getFilePathReference()
            logPath = outObjD["dst3"].getFilePathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.setDebugMode(flag=True)
            # dp.imp(inpMapPath)
            #  add any extra command line options
            options = str(uD["options"])
            if (options is not None) and (len(options) > 0):
                dp.addInput(name="options", value=options)

            dp.addInput(name="input_map_file_path", value=inpMapPath)
            dp.addInput(name="output_map_file_path", value=outMapPath)
            dp.op("deposit-update-map-header-in-place")
            # dp.op("annot-update-map-header-in-place")
            dp.exp(rptPath)
            dp.expLog(logPath)

            if self._verbose:
                self._lfh.write("+EmUtils.mapFixInPlaceOp() - input map  file path:  %s\n" % inpMapPath)
                self._lfh.write("+EmUtils.mapFixInPlaceOp() - output map file path:  %s\n" % outMapPath)
                self._lfh.write("+EmUtils.mapFixInPlaceOp() - report file path:      %s\n" % rptPath)
                self._lfh.write("+EmUtils.mapFixInPlaceOp() - log file path:         %s\n" % logPath)
                self._lfh.write("+EmUtils.mapFixInPlaceOp() - command option:        %s\n" % options)

            if self.__cleanUp:
                dp.cleanup()
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def em2emSpiderOp(self, **kwArgs):
        """Run em2em to convert Spider map to CCP4 format"""
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            inpMapPath = inpObjD["src"].getFilePathReference()
            outMapPath = outObjD["dst1"].getFilePathReference()
            dirPath = outObjD["dst1"].getDirPathReference()
            logPath = outObjD["dst2"].getFilePathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(inpMapPath)
            #  add any extra command line options
            pixelSpacingX = 1.0
            pixelSpacingY = 1.0
            pixelSpacingZ = 1.0
            pixelSpacingX = str(uD["pixel-spacing-x"])
            pixelSpacingY = str(uD["pixel-spacing-y"])
            pixelSpacingZ = str(uD["pixel-spacing-z"])
            dp.addInput(name="pixel-spacing-x", value=pixelSpacingX)
            dp.addInput(name="pixel-spacing-y", value=pixelSpacingY)
            dp.addInput(name="pixel-spacing-z", value=pixelSpacingZ)
            #
            dp.op("em2em-spider")
            dp.expLog(logPath)
            dp.exp(outMapPath)

            if self._verbose:
                self._lfh.write("+EmUtils.em2emSpiderOp() - input map  file path:  %s\n" % inpMapPath)
                self._lfh.write("+EmUtils.em2emSpiderOp() - output map file path:  %s\n" % outMapPath)
                self._lfh.write("+EmUtils.em2emSpiderOp() - log file path:         %s\n" % logPath)

            if self.__cleanUp:
                dp.cleanup()
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def mapFixInPlaceAltOp(self, **kwArgs):
        """MapFix operation on input map (in place).

        - with automatic input and output map file selection -

        """
        try:
            (inpObjD, outObjD, _uD, _pD) = self._getArgs(kwArgs)
            inpMapPath = inpObjD["src1"].getFilePathReference()
            inpArgsPath = inpObjD["src2"].getFilePathReference()
            outMapPath = outObjD["dst1"].getFilePathReference()
            dirPath = outObjD["dst1"].getDirPathReference()
            rptPath = outObjD["dst2"].getFilePathReference()
            logPath = outObjD["dst3"].getFilePathReference()
            #
            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.setDebugMode(flag=True)
            #
            #  add any extra command line options
            options = None
            try:
                ifh = open(inpArgsPath, "r")
                options = ifh.read()
                ifh.close()
                dp.addInput(name="options", value=options)
            except Exception as _e:  # noqa: F841
                pass
            dp.addInput(name="input_map_file_path", value=inpMapPath)
            dp.addInput(name="output_map_file_path", value=outMapPath)
            dp.op("annot-update-map-header-in-place")
            dp.exp(rptPath)
            dp.expLog(logPath)

            if self._verbose:
                self._lfh.write("+EmUtils.mapFixOp() - input map  file path:         %s\n" % inpMapPath)
                self._lfh.write("+EmUtils.mapFixOp() - command line args file path:  %s\n" % inpArgsPath)
                self._lfh.write("+EmUtils.mapFixOp() - output map file path:         %s\n" % outMapPath)
                self._lfh.write("+EmUtils.mapFixOp() - report file path:             %s\n" % rptPath)
                self._lfh.write("+EmUtils.mapFixOp() - log file path:                %s\n" % logPath)
                self._lfh.write("+EmUtils.mapFixOp() - command option:               %s\n" % options)

            if self.__cleanUp:
                dp.cleanup()
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def mapFixInPlaceCfgOp(self, **kwArgs):
        """execution wrapper for mapfix operations for all maps of a particular type.

        --- using explicit data file selection via configuraton file selection.

            map-content-type = one of em-volume, em-mask-volume, em-half-volume, em-additional-volume
            partlist = [1,2,3]
            cmdline-arg-list = ['-voxel ...   ','-voxel .... ','-voxel .... ']



        """
        try:
            (inpObjD, _outObjD, _uD, _pD) = self._getArgs(kwArgs)
            inpCfgPath = inpObjD["src1"].getFilePathReference()
            dataSetId = inpObjD["src1"].getDepositionDataSetId()
            dirPath = inpObjD["src1"].getDirPathReference()
            storageType = inpObjD["src1"].getStorageType()
            #
            #  Read data configuration information ---
            arg = None
            cTupL = []
            try:
                cD = json.load(open(inpCfgPath, "r"))
                pL = cD["part-list"]
                inpMileStone = cD["map-content-milestone"] if "map-content-milestone" in cD else None
                mapContentType = cD["map-content-type"]
                cmdLineArgList = cD["cmd-line-arg-list"]
                for p, arg in zip(pL, cmdLineArgList):
                    cTupL.append((p, mapContentType, arg))
            except Exception as _e:  # noqa: F841
                self._lfh.write("+EmUtils.mapFixInPlaceCfgOp() - failed processing configuration file  %s\n" % inpCfgPath)
                traceback.print_exc(file=self._lfh)

            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            pI = PathInfo(siteId=siteId, verbose=self._verbose, log=self._lfh)

            for p, mT, arg in cTupL:
                inpMapPath = pI.getFilePath(dataSetId, contentType=mT, formatType="map", fileSource=storageType, versionId="latest", partNumber=p, mileStone=inpMileStone)
                outMapPath = pI.getFilePath(dataSetId, contentType=mT, formatType="map", fileSource=storageType, versionId="next", partNumber=p, mileStone=None)
                rptPath = pI.getFilePath(dataSetId, contentType="mapfix-header-report", formatType="json", fileSource=storageType, versionId="next", partNumber=p, mileStone=None)
                logPath = pI.getFilePath(dataSetId, contentType="mapfix-report", formatType="txt", fileSource=storageType, versionId="next", partNumber=p, mileStone=None)

                dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
                dp.setDebugMode(flag=True)
                dp.addInput(name="options", value=arg)
                dp.addInput(name="input_map_file_path", value=inpMapPath)
                dp.addInput(name="output_map_file_path", value=outMapPath)
                dp.op("deposit-update-map-header-in-place")
                dp.exp(rptPath)
                dp.expLog(logPath)

                if self._verbose:
                    self._lfh.write("+EmUtils.mapFixInPlaceCfgOp() - config file path:             %s\n" % inpCfgPath)
                    self._lfh.write("+EmUtils.mapFixInPlaceCfgOp() - input map  file path:         %s\n" % inpMapPath)
                    self._lfh.write("+EmUtils.mapFixInPlaceCfgOp() - output map file path:         %s\n" % outMapPath)
                    self._lfh.write("+EmUtils.mapFixInPlaceCfgOp() - report file path:             %s\n" % rptPath)
                    self._lfh.write("+EmUtils.mapFixInPlaceCfgOp() - log file path:                %s\n" % logPath)
                    self._lfh.write("+EmUtils.mapFixInPlaceCfgOp() - command option:               %s\n" % arg)

                if self.__cleanUp:
                    dp.cleanup()
            return True
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return False

    def imageMagickOp(self, **kwArgs):
        """
        Run ImageMagick to convert uploaded image to 400x400 (for atlas pages)

        """
        try:
            (inpObjD, outObjD, uD, _pD) = self._getArgs(kwArgs)
            inpImgPath = inpObjD["src"].getFilePathReference()
            outImgPath = outObjD["dst1"].getFilePathReference()
            dirPath = outObjD["dst1"].getDirPathReference()
            logPath = outObjD["dst2"].getFilePathReference()

            cI = ConfigInfo()
            siteId = cI.get("SITE_PREFIX")
            dp = RcsbDpUtility(tmpPath=dirPath, siteId=siteId, verbose=self._verbose, log=self._lfh)
            dp.imp(inpImgPath)

            #  add any extra command line options
            options = str(uD["options"])
            if (options is not None) and (len(options) > 0):
                dp.addInput(name="options", value=options)

            ret = dp.op("img-convert")
            dp.expLog(logPath)
            dp.exp(outImgPath)

            if self._verbose:
                self._lfh.write("+EmUtils.imageMagickOp() - input image  file path:  %s\n" % inpImgPath)
                self._lfh.write("+EmUtils.imageMagickOp() - output image file path:  %s\n" % outImgPath)
                self._lfh.write("+EmUtils.imageMagickOp() - log file path:         %s\n" % logPath)

            if self.__cleanUp:
                dp.cleanup()

            return ret
        except Exception as _e:  # noqa: F841
            traceback.print_exc(file=self._lfh)
            return -100
